"""Utilities for validating estimators."""

__all__ = [
    "is_sklearn_estimator",
    "is_sklearn_classifier",
    "is_sklearn_regressor",
    "is_sklearn_clusterer",
]

from aeon.base import BaseEstimator as AeonBaseEstimator
from sklearn.base import BaseEstimator as SklearnBaseEstimator
from sklearn.base import is_classifier, is_regressor
from tsml.base import BaseTimeSeriesEstimator
from tsml.utils.validation import is_clusterer


def is_sklearn_estimator(estimator):
    """Check if estimator is a scikit-learn estimator."""
    return (
        isinstance(estimator, SklearnBaseEstimator)
        and not isinstance(estimator, AeonBaseEstimator)
        and not isinstance(estimator, BaseTimeSeriesEstimator)
    )


def is_sklearn_classifier(classifier):
    """Check if estimator is a scikit-learn classifier."""
    return is_sklearn_estimator(classifier) and is_classifier(classifier)


def is_sklearn_regressor(regressor):
    """Check if estimator is a scikit-learn regressor."""
    return is_sklearn_estimator(regressor) and is_regressor(regressor)


def is_sklearn_clusterer(clusterer):
    """Check if estimator is a scikit-learn clusterer."""
    return is_sklearn_estimator(clusterer) and is_clusterer(clusterer)



"""Tests for classification experiments."""

__author__ = ["MatthewMiddlehurst"]

import os
import runpy

import pytest
from aeon.registry import all_estimators
from tsml.dummy import DummyRegressor

from tsml_eval.datasets._test_data._data_sizes import DATA_TEST_SIZES
from tsml_eval.experiments import (
    classification_experiments,
    get_classifier_by_name,
    run_classification_experiment,
    set_classifier,
    threaded_classification_experiments,
)
from tsml_eval.experiments.tests import _CLASSIFIER_RESULTS_PATH
from tsml_eval.testing.testing_utils import (
    _TEST_DATA_PATH,
    _check_set_method,
    _check_set_method_results,
)
from tsml_eval.utils.tests.test_results_writing import _check_classification_file_format


@pytest.mark.parametrize(
    "classifier",
    ["DummyClassifier-tsml", "DummyClassifier-aeon", "DummyClassifier-sklearn"],
)
@pytest.mark.parametrize(
    "dataset",
    ["MinimalChinatown", "UnequalMinimalChinatown", "EqualMinimalJapaneseVowels"],
)
def test_run_classification_experiment(classifier, dataset):
    """Test classification experiments with test data and classifier."""
    args = [
        _TEST_DATA_PATH,
        _CLASSIFIER_RESULTS_PATH,
        classifier,
        dataset,
        "0",
        "-tr",
    ]

    classification_experiments.run_experiment(args)

    test_file = (
        f"{_CLASSIFIER_RESULTS_PATH}{classifier}/Predictions/{dataset}/"
        "testResample0.csv"
    )
    train_file = (
        f"{_CLASSIFIER_RESULTS_PATH}{classifier}/Predictions/{dataset}/"
        "trainResample0.csv"
    )

    assert os.path.exists(test_file) and os.path.exists(train_file)

    _check_classification_file_format(
        test_file, num_results_lines=DATA_TEST_SIZES[dataset]
    )
    _check_classification_file_format(
        train_file, num_results_lines=DATA_TEST_SIZES[dataset]
    )

    # test present results checking
    classification_experiments.run_experiment(args)

    os.remove(test_file)
    os.remove(train_file)


def test_run_classification_experiment_main():
    """Test classification experiments main with test data and classifier."""
    classifier = "ROCKET"
    dataset = "MinimalChinatown"

    # run twice to test results present check
    for _ in range(2):
        runpy.run_path(
            (
                "./tsml_eval/experiments/classification_experiments.py"
                if os.getcwd().split("\\")[-1] != "tests"
                else "../classification_experiments.py"
            ),
            run_name="__main__",
        )

    test_file = (
        f"{_CLASSIFIER_RESULTS_PATH}{classifier}/Predictions/{dataset}/"
        "testResample0.csv"
    )
    assert os.path.exists(test_file)
    _check_classification_file_format(test_file)

    os.remove(test_file)


def test_run_threaded_classification_experiment():
    """Test threaded classification experiments with test data and classifier."""
    classifier = "ROCKET"
    dataset = "MinimalChinatown"

    args = [
        _TEST_DATA_PATH,
        _CLASSIFIER_RESULTS_PATH,
        classifier,
        dataset,
        "1",
        "-nj",
        "2",
        # also test normalisation and benchmark time here
        "--row_normalise",
        "--benchmark_time",
    ]

    threaded_classification_experiments.run_experiment(args)

    test_file = (
        f"{_CLASSIFIER_RESULTS_PATH}{classifier}/Predictions/{dataset}/"
        "testResample1.csv"
    )
    assert os.path.exists(test_file)
    _check_classification_file_format(test_file)

    # test present results checking
    threaded_classification_experiments.run_experiment(args)

    # this covers the main method and experiment function result file checking
    runpy.run_path(
        (
            "./tsml_eval/experiments/threaded_classification_experiments.py"
            if os.getcwd().split("\\")[-1] != "tests"
            else "../threaded_classification_experiments.py"
        ),
        run_name="__main__",
    )

    os.remove(test_file)


def test_run_classification_experiment_invalid_build_settings():
    """Test run_classification_experiment method with invalid build settings."""
    with pytest.raises(ValueError, match="Both test_file and train_file"):
        run_classification_experiment(
            [],
            [],
            [],
            [],
            None,
            "",
            build_test_file=False,
            build_train_file=False,
        )


def test_run_classification_experiment_invalid_estimator():
    """Test run_classification_experiment method with invalid estimator."""
    with pytest.raises(TypeError, match="classifier must be a"):
        run_classification_experiment(
            [],
            [],
            [],
            [],
            DummyRegressor(),
            "",
        )


def test_get_classifier_by_name():
    """Test get_classifier_by_name method."""
    classifier_lists = [
        set_classifier.convolution_based_classifiers,
        set_classifier.deep_learning_classifiers,
        set_classifier.dictionary_based_classifiers,
        set_classifier.distance_based_classifiers,
        set_classifier.feature_based_classifiers,
        set_classifier.hybrid_classifiers,
        set_classifier.interval_based_classifiers,
        set_classifier.other_classifiers,
        set_classifier.shapelet_based_classifiers,
        set_classifier.vector_classifiers,
    ]

    classifier_dict = {}
    all_classifier_names = []

    for classifier_list in classifier_lists:
        _check_set_method(
            get_classifier_by_name,
            classifier_list,
            classifier_dict,
            all_classifier_names,
        )

    _check_set_method_results(
        classifier_dict, estimator_name="Classifiers", method_name="set_classifier"
    )


def test_get_classifier_by_name_invalid():
    """Test get_classifier_by_name method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN CLASSIFIER"):
        get_classifier_by_name("invalid")


def test_aeon_classifiers_available():
    """Test all aeon classifiers are available."""
    excluded = [
        # composable
        "ChannelEnsembleClassifier",
        "ClassifierPipeline",
        "WeightedEnsembleClassifier",
        # just missing
        "IndividualLITEClassifier",
        "OrdinalTDE",
        "IndividualOrdinalTDE",
        "IntervalForestClassifier",
        "SupervisedIntervalClassifier",
        "LearningShapeletClassifier",
    ]

    est = [e for e, _ in all_estimators(estimator_types="classifier")]
    for e in est:
        if e in excluded:
            continue

        try:
            assert get_classifier_by_name(e) is not None
        except ModuleNotFoundError:
            continue



"""Miscellaneous functions for tsml_eval."""

__all__ = [
    "str_in_nested_list",
    "pair_list_to_dict",
    "time_to_milliseconds",
    "rank_array",
]

import numpy as np


def str_in_nested_list(nested_list, item):
    """Find an item in a nested list."""
    if item in (s.casefold() for s in nested_list if isinstance(s, str)):
        return True
    else:
        return any(
            str_in_nested_list(nl, item) for nl in nested_list if isinstance(nl, list)
        )


def pair_list_to_dict(pl):
    """Convert a 2d list of pairs to a dict.

    Each list item must be a tuple or list of length 2. The first item in each pair
    is used as the key, the second as the value.

    If ls is None, returns an empty dict.
    """
    return {} if pl is None else {k: v for k, v in pl}


def time_to_milliseconds(time_value, time_unit):
    """Convert a time value from the given time unit to milliseconds.

    Parameters
    ----------
    time_value : float
        The time value to convert.
    time_unit : str {"nanoseconds", "microseconds", "milliseconds", "seconds",
                    "minutes", "hours", "days"}
        The current time unit of the value.

    Returns
    -------
    float
        The time in milliseconds.
    """
    time_unit = time_unit.lower().strip()
    time_units = {
        "nanoseconds": 1e-6,
        "microseconds": 1e-3,
        "milliseconds": 1,
        "seconds": 1e3,
        "minutes": 60e3,
        "hours": 3600e3,
        "days": 86400e3,
    }

    if time_unit not in time_units:
        raise ValueError(f"Unknown time unit: {time_unit}")

    # Convert the time value to milliseconds
    return time_value * time_units[time_unit]


def rank_array(arr, higher_better=True):
    """
    Assign a rank to each value in a 1D numpy array.

    A lower rank number is assumed to be better. Lower values can receive better ranks
    or vice versa based on the `higher_better` parameter. Equal values receive the
    average of the ranks they would cover.

    Parameters
    ----------
    arr : numpy.ndarray
        The input 1D array containing values to be ranked.
    higher_better : bool, default=True
        If True, lower values receive better ranks.
        If False (default), higher values receive better ranks.

    Returns
    -------
    ranks : numpy.ndarray
        Array of ranks, same shape as `arr`.
    """
    # argsort returns indices that would sort the array
    sorter = np.argsort(arr)
    ranks = np.zeros(len(arr), dtype=float)
    ranks[sorter] = np.arange(1, len(arr) + 1)

    # Handle ties: find unique values and their corresponding indices
    unique_vals, inv_sorter = np.unique(arr, return_inverse=True)
    for i in np.unique(inv_sorter):
        ranks[inv_sorter == i] = np.mean(ranks[inv_sorter == i])

    if higher_better:
        ranks = len(arr) + 1 - ranks

    return ranks



"""Tests for the results IO functionality."""

import os

from tsml_eval.evaluation.storage.classifier_results import ClassifierResults
from tsml_eval.evaluation.storage.clusterer_results import ClustererResults
from tsml_eval.evaluation.storage.forecaster_results import ForecasterResults
from tsml_eval.evaluation.storage.regressor_results import RegressorResults
from tsml_eval.testing.testing_utils import _TEST_OUTPUT_PATH, _TEST_RESULTS_PATH
from tsml_eval.utils.results_validation import validate_results_file


def test_classifier_results():
    """Test ClassifierResults loading and saving."""
    cr = ClassifierResults().load_from_file(
        _TEST_RESULTS_PATH
        + "/classification/ROCKET/Predictions/MinimalChinatown/testResample0.csv"
    )
    cr.save_to_file(_TEST_OUTPUT_PATH + "/classification/results_io/")

    assert validate_results_file(
        _TEST_OUTPUT_PATH + "/classification/results_io/testResample0.csv"
    )

    os.remove(_TEST_OUTPUT_PATH + "/classification/results_io/testResample0.csv")


def test_java_classifier_results():
    """Test ClassifierResults loading and saving."""
    cr = ClassifierResults().load_from_file(
        _TEST_RESULTS_PATH + "/classification/javaResultsFile.csv"
    )
    cr.save_to_file(_TEST_OUTPUT_PATH + "/classification/results_io/java/")

    assert validate_results_file(
        _TEST_OUTPUT_PATH + "/classification/results_io/java/testResample0.csv"
    )

    os.remove(_TEST_OUTPUT_PATH + "/classification/results_io/java/testResample0.csv")


def test_clusterer_results():
    """Test ClustererResults loading and saving."""
    cr = ClustererResults().load_from_file(
        _TEST_RESULTS_PATH
        + "/clustering/KMeans/Predictions/MinimalChinatown/trainResample0.csv"
    )
    cr.save_to_file(_TEST_OUTPUT_PATH + "/clustering/results_io/")

    assert validate_results_file(
        _TEST_OUTPUT_PATH + "/clustering/results_io/trainResample0.csv"
    )

    os.remove(_TEST_OUTPUT_PATH + "/clustering/results_io/trainResample0.csv")


def test_regressor_results():
    """Test RegressorResults loading and saving."""
    cr = RegressorResults().load_from_file(
        _TEST_RESULTS_PATH
        + "/regression/ROCKET/Predictions/MinimalGasPrices/testResample0.csv"
    )
    cr.save_to_file(_TEST_OUTPUT_PATH + "/regression/results_io/")

    assert validate_results_file(
        _TEST_OUTPUT_PATH + "/regression/results_io/testResample0.csv"
    )

    os.remove(_TEST_OUTPUT_PATH + "/regression/results_io/testResample0.csv")


def test_forecaster_results():
    """Test ForecasterResults loading and saving."""
    cr = ForecasterResults().load_from_file(
        _TEST_RESULTS_PATH
        + "/forecasting/NaiveForecaster/Predictions/ShampooSales/testResample0.csv"
    )
    cr.save_to_file(_TEST_OUTPUT_PATH + "/forecasting/results_io/")

    assert validate_results_file(
        _TEST_OUTPUT_PATH + "/forecasting/results_io/testResample0.csv"
    )

    os.remove(_TEST_OUTPUT_PATH + "/forecasting/results_io/testResample0.csv")



"""Tests for regression experiments."""

__author__ = ["MatthewMiddlehurst"]

import os
import runpy

import pytest
from aeon.registry import all_estimators
from tsml.dummy import DummyClassifier

from tsml_eval.datasets._test_data._data_sizes import DATA_TEST_SIZES
from tsml_eval.experiments import (
    get_regressor_by_name,
    regression_experiments,
    run_regression_experiment,
    set_regressor,
    threaded_regression_experiments,
)
from tsml_eval.experiments.tests import _REGRESSOR_RESULTS_PATH
from tsml_eval.testing.testing_utils import (
    _TEST_DATA_PATH,
    _check_set_method,
    _check_set_method_results,
)
from tsml_eval.utils.tests.test_results_writing import _check_regression_file_format


@pytest.mark.parametrize(
    "regressor",
    ["DummyRegressor-tsml", "DummyRegressor-aeon", "DummyRegressor-sklearn"],
)
@pytest.mark.parametrize(
    "dataset",
    ["MinimalGasPrices", "UnequalMinimalGasPrices", "MinimalCardanoSentiment"],
)
def test_run_regression_experiment(regressor, dataset):
    """Test regression experiments with test data and regressor."""
    if regressor == "DummyRegressor-aeon" and dataset == "UnequalMinimalGasPrices":
        return  # todo remove when aeon dummy supports unequal

    args = [
        _TEST_DATA_PATH,
        _REGRESSOR_RESULTS_PATH,
        regressor,
        dataset,
        "0",
        "-tr",
    ]

    regression_experiments.run_experiment(args)

    test_file = (
        f"{_REGRESSOR_RESULTS_PATH}{regressor}/Predictions/{dataset}/testResample0.csv"
    )
    train_file = (
        f"{_REGRESSOR_RESULTS_PATH}{regressor}/Predictions/{dataset}/trainResample0.csv"
    )

    assert os.path.exists(test_file) and os.path.exists(train_file)

    _check_regression_file_format(test_file, num_results_lines=DATA_TEST_SIZES[dataset])
    _check_regression_file_format(
        train_file, num_results_lines=DATA_TEST_SIZES[dataset]
    )

    # test present results checking
    regression_experiments.run_experiment(args)

    os.remove(test_file)
    os.remove(train_file)


def test_run_regression_experiment_main():
    """Test regression experiments main with test data and regressor."""
    regressor = "ROCKET"
    dataset = "MinimalGasPrices"

    # run twice to test results present check
    for _ in range(2):
        runpy.run_path(
            (
                "./tsml_eval/experiments/regression_experiments.py"
                if os.getcwd().split("\\")[-1] != "tests"
                else "../regression_experiments.py"
            ),
            run_name="__main__",
        )

    test_file = (
        f"{_REGRESSOR_RESULTS_PATH}{regressor}/Predictions/{dataset}/testResample0.csv"
    )
    assert os.path.exists(test_file)
    _check_regression_file_format(test_file)

    os.remove(test_file)


def test_run_threaded_regression_experiment():
    """Test threaded regression experiments with test data and regressor."""
    regressor = "ROCKET"
    dataset = "MinimalGasPrices"

    args = [
        _TEST_DATA_PATH,
        _REGRESSOR_RESULTS_PATH,
        regressor,
        dataset,
        "1",
        "-nj",
        "2",
        # also test normalisation and benchmark time here
        "--row_normalise",
        "--benchmark_time",
    ]

    threaded_regression_experiments.run_experiment(args)

    test_file = (
        f"{_REGRESSOR_RESULTS_PATH}{regressor}/Predictions/{dataset}/testResample1.csv"
    )
    assert os.path.exists(test_file)
    _check_regression_file_format(test_file)

    # test present results checking
    threaded_regression_experiments.run_experiment(args)

    # this covers the main method and experiment function result file checking
    runpy.run_path(
        (
            "./tsml_eval/experiments/threaded_regression_experiments.py"
            if os.getcwd().split("\\")[-1] != "tests"
            else "../threaded_regression_experiments.py"
        ),
        run_name="__main__",
    )

    os.remove(test_file)


def test_run_regression_experiment_invalid_build_settings():
    """Test run_regression_experiment method with invalid build settings."""
    with pytest.raises(ValueError, match="Both test_file and train_file"):
        run_regression_experiment(
            [],
            [],
            [],
            [],
            None,
            "",
            build_test_file=False,
            build_train_file=False,
        )


def test_run_regression_experiment_invalid_estimator():
    """Test run_regression_experiment method with invalid estimator."""
    with pytest.raises(TypeError, match="regressor must be a"):
        run_regression_experiment(
            [],
            [],
            [],
            [],
            DummyClassifier(),
            "",
        )


def test_get_regressor_by_name():
    """Test get_regressor_by_name method."""
    regressor_lists = [
        set_regressor.convolution_based_regressors,
        set_regressor.deep_learning_regressors,
        set_regressor.distance_based_regressors,
        set_regressor.feature_based_regressors,
        set_regressor.hybrid_regressors,
        set_regressor.interval_based_regressors,
        set_regressor.other_regressors,
        set_regressor.shapelet_based_regressors,
        set_regressor.vector_regressors,
    ]

    regressor_dict = {}
    all_regressor_names = []

    for regressor_list in regressor_lists:
        _check_set_method(
            get_regressor_by_name,
            regressor_list,
            regressor_dict,
            all_regressor_names,
        )

    _check_set_method_results(
        regressor_dict, estimator_name="Regressors", method_name="set_regressor"
    )


def test_get_regressor_by_name_invalid():
    """Test get_regressor_by_name method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN REGRESSOR"):
        get_regressor_by_name("invalid")


def test_aeon_regressors_available():
    """Test all aeon regressors are available."""
    excluded = [
        # composable
        "RegressorPipeline",
        # just missing
        "IndividualLITERegressor",
        "IntervalForestRegressor",
    ]

    est = [e for e, _ in all_estimators(estimator_types="regressor")]
    for e in est:
        if e in excluded:
            continue

        try:
            assert get_regressor_by_name(e) is not None
        except ModuleNotFoundError:
            continue



"""Hierarchical Vote Collective of Transformation-based Ensembles (HIVE-COTE) from file.

Hybrid ensemble of classifiers from separate time series classification
representations, using the weighted probabilistic CAWPE as an ensemble controller.
"""

__author__ = ["MatthewMiddlehurst", "ander-hg"]
__all__ = ["FromFileHIVECOTE"]

import numpy as np
from aeon.classification import BaseClassifier
from sklearn import preprocessing
from sklearn.metrics import accuracy_score
from sklearn.model_selection import StratifiedKFold
from sklearn.utils import check_random_state


class FromFileHIVECOTE(BaseClassifier):
    """HIVE-COTE from file.

    Builds the Hierarchical Vote Collective of Transformation-based Ensembles
    (HIVE-COTE) from results files.
    For example, HC2 is n ensemble of the STC, DrCIF, Arsenal and TDE classifiers from
    different feature representations using the CAWPE structure as described in [1].

    Parameters
    ----------
    classifiers : list
        The paths to results files used. i.e. Arsenal, DrCIF, STC and TDE files
        for HC2.
    alpha : int, default=4
        The exponent to extenuate differences in classifiers and weighting with the
        accuracy estimate.
    tune_alpha : bool, default=False
        Tests alpha [1..10] and uses the best value.
    new_weights : list, default=None
        The list of weights to use. Skips finding of weights if not None.
    acc_filter : (str, float) tuple, default=None
        Removes the worst component based or on accuracies on
        training data if first item is "train" or test data if first item is "test".
        Be warned that using the test data in this way is cheating, and should only be
        used for exploration.
        The second item is the threshold for the filter, e.g. 0.5 will remove all
        components with accuracies less than 50% of the best component.
        Will always keep the most accurate component.
    overwrite_y : bool, default=False
        If True, the labels in the loaded files will overwrite the labels
        passed in the fit method.
    skip_y_check : bool, default=False
        If True, the labels in the loaded files will not be checked against
        the labels passed in the fit method.
    skip_shape_check : bool, default=False
        If True, the shape of the loaded files (n_cases and n_classes) will not be
        checked against the shape of the data passed in the fit method. This can cause
        breakage, so use with caution.
    random_state : int or None, default=None
        Seed for random number generation.

    Attributes
    ----------
    weights_ : list
        The weight for Arsenal, DrCIF, STC and TDE probabilities.

    References
    ----------
    .. [1] Middlehurst, Matthew, James Large, Michael Flynn, Jason Lines, Aaron Bostrom,
       and Anthony Bagnall. "HIVE-COTE 2.0: a new meta ensemble for time series
       classification." Machine Learning (2021).
    """

    _tags = {
        "capability:multivariate": True,
        "algorithm_type": "hybrid",
    }

    def __init__(
        self,
        classifiers,
        alpha=4,
        tune_alpha=False,
        new_weights=None,
        acc_filter=None,
        overwrite_y=False,
        skip_y_check=False,
        skip_shape_check=False,
        random_state=None,
    ):
        self.classifiers = classifiers
        self.alpha = alpha
        self.tune_alpha = tune_alpha
        self.new_weights = new_weights
        self.acc_filter = acc_filter
        self.overwrite_y = overwrite_y
        self.skip_y_check = skip_y_check
        self.skip_shape_check = skip_shape_check
        self.random_state = random_state

        super().__init__()

    def _fit(self, X, y):
        self.weights_ = []

        if self.new_weights:
            acc_list = self.new_weights
        else:
            n_instances, _, _ = X.shape
            acc_list = []

            # load train file at path (trainResample.csv if random_state is None,
            # trainResample{self.random_state}.csv otherwise)
            if self.random_state is not None:
                file_name = f"trainResample{self.random_state}.csv"
            else:
                file_name = "trainResample.csv"

            if self.tune_alpha:
                X_probas = np.zeros(
                    (n_instances, len(self.classifiers), self.n_classes_)
                )

            for i, path in enumerate(self.classifiers):
                f = open(path + file_name)
                lines = f.readlines()
                line2 = lines[2].split(",")

                # verify file matches data
                if not self.skip_shape_check:
                    if len(lines) - 3 != n_instances:
                        raise ValueError(
                            f"n_instances of {path + file_name} does not match X, "
                            f"expected {X.shape[0]}, got {len(lines) - 3}"
                        )
                    if not self.skip_y_check and self.n_classes_ != int(line2[5]):
                        raise ValueError(
                            f"n_classes of {path + file_name} does not match X, "
                            f"expected {len(np.unique(y))}, got {line2[5]}"
                        )

                for j in range(n_instances):
                    line = lines[j + 3].split(",")

                    if self.overwrite_y:
                        if i == 0:
                            y[j] = float(line[0])
                        elif not self.skip_y_check:
                            assert y[j] == float(line[0])
                    elif not self.skip_y_check:
                        if i == 0:
                            le = preprocessing.LabelEncoder()
                            y = le.fit_transform(y)
                        assert float(line[0]) == y[j]

                    if self.tune_alpha:
                        X_probas[j][i] = [float(k) for k in (line[3:])]

                acc_list.append(float(line2[0]))

            self._alpha = (
                self._tune_alpha(X_probas, y) if self.tune_alpha else self.alpha
            )

        # add a weight to the weight list based on the files accuracy
        for acc in acc_list:
            self.weights_.append(acc**self._alpha)

        self._use_classifier = [True for _ in range(len(self.classifiers))]
        if self.acc_filter is not None:
            self._acc_filter = (
                ("train", self.acc_filter)
                if isinstance(self.acc_filter, float)
                else self.acc_filter
            )

            if self._acc_filter[0] != "train" and self._acc_filter[0] != "test":
                raise ValueError(
                    f"acc_filter[0] must be 'train' or 'test', got "
                    f"{self._acc_filter[0]}"
                )
            elif self._acc_filter[1] <= 0 or self._acc_filter[1] >= 1:
                raise ValueError(
                    "acc_filter[1] must be in between 0 and 1, got "
                    f"{self._acc_filter[1]}"
                )

            if self._acc_filter[0] == "test":
                # load test file at path (testResample.csv if random_state is None,
                # testResample{self.random_state}.csv otherwise)
                if self.random_state is not None:
                    file_name = f"testResample{self.random_state}.csv"
                else:
                    file_name = "testResample.csv"

                acc_list = []
                for path in self.classifiers:
                    f = open(path + file_name)
                    lines = f.readlines()
                    line2 = lines[2].split(",")

                    acc_list.append(float(line2[0]))

            argmax = np.argmax(acc_list)

            # add a weight to the weight list based on the files accuracy
            for i, acc in enumerate(acc_list):
                if i != argmax and acc < acc_list[argmax] * self._acc_filter[1]:
                    self._use_classifier[i] = False

    def _predict(self, X):
        rng = check_random_state(self.random_state)
        return np.array(
            [
                self.classes_[int(rng.choice(np.flatnonzero(prob == prob.max())))]
                for prob in self.predict_proba(X)
            ]
        )

    def _predict_proba(self, X):
        n_instances, _, _ = X.shape

        # load test file at path (testResample.csv if random_state is None,
        # testResample{self.random_state}.csv otherwise)
        if self.random_state is not None:
            file_name = f"testResample{self.random_state}.csv"
        else:
            file_name = "testResample.csv"

        dists = np.zeros((n_instances, self.n_classes_))

        if not self.skip_y_check:
            y = np.zeros(n_instances)

        for i, path in enumerate(self.classifiers):
            f = open(path + file_name)
            lines = f.readlines()
            line2 = lines[2].split(",")

            # verify file matches data
            if not self.skip_shape_check:
                if len(lines) - 3 != n_instances:
                    raise ValueError(
                        f"n_instances of {path + file_name} does not match X, "
                        f"expected {X.shape[0]}, got {len(lines) - 3}"
                    )
                if not self.skip_y_check and self.n_classes_ != int(line2[5]):
                    raise ValueError(
                        f"n_classes of {path + file_name} does not match X, "
                        f"expected {self.n_classes_}, got {line2[5]}"
                    )

            # apply this files weights to the probabilities in the test file
            for j in range(n_instances):
                line = lines[j + 3].split(",")

                if not self.skip_y_check:
                    if i == 0:
                        y[j] = float(line[0])
                    else:
                        assert y[j] == float(line[0])

                if self._use_classifier[i]:
                    dists[j] = np.add(
                        dists[j],
                        [float(k) for k in (line[3:])]
                        * (np.ones(self.n_classes_) * self.weights_[i]),
                    )

        # Make each instances probability array sum to 1 and return
        return dists / dists.sum(axis=1, keepdims=True)

    def _tune_alpha(self, X_probas, y):
        n_instances = len(y)
        n_files = X_probas.shape[1]

        n_splits = 10
        _, counts = np.unique(y, return_counts=True)
        min_class = np.min(counts)
        if min_class < n_splits:
            n_splits = min_class

        if n_splits == 1:
            return self.alpha

        alpha_values = range(1, 10)  # tested alpha values
        avg_alpha_acc = np.zeros(len(alpha_values))  # performance of each alpha value
        for i, alpha in enumerate(alpha_values):
            kf = StratifiedKFold(
                n_splits=n_splits,
                shuffle=True,
                random_state=(
                    np.random.randint(0, np.iinfo(np.int32).max)
                    if self.random_state is None
                    else self.random_state
                ),
            )
            preds = np.zeros(n_instances)

            for train_index, test_index in kf.split(X_probas, y):
                weight_list = []
                for n in range(n_files):
                    train_preds = np.argmax(X_probas[train_index, n], axis=1)
                    train_acc = accuracy_score(y[train_index], train_preds)
                    weight_list.append(train_acc**alpha)

                dists = np.zeros((len(test_index), self.n_classes_))
                # apply the weights to the probabilities in the test set
                for n in range(n_files):
                    for v in range(len(test_index)):
                        dists[v] = np.add(
                            dists[v],
                            [k for k in X_probas[test_index[v], n]]
                            * (np.ones(self.n_classes_) * weight_list[n]),
                        )

                # Make each instances probability array sum to 1
                dists = dists / dists.sum(axis=1, keepdims=True)
                preds[test_index] = np.argmax(dists, axis=1)

            avg_alpha_acc[i] = accuracy_score(y, preds)

        best_alpha = alpha_values[avg_alpha_acc.argmax()]

        return best_alpha

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.
            For classifiers, a "default" set of parameters should be provided for
            general testing, and a "results_comparison" set for comparing against
            previously recorded results if the general set does not produce suitable
            probabilities to compare against.

        Returns
        -------
        params : dict or list of dict, default={}
            Parameters to create testing instances of the class.
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`.
        """
        from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH

        file_paths = [
            _TEST_RESULTS_PATH + "/classification/TestResults/Test1/",
            _TEST_RESULTS_PATH + "/classification/TestResults/Test2/",
        ]
        return {
            "classifiers": file_paths,
            "skip_y_check": True,
            "skip_shape_check": True,
            "random_state": 0,
        }



"""Test estimators implemented in tsml-eval."""

from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from tsml.utils.testing import parametrize_with_checks

from tsml_eval.estimators import (
    SklearnToTsmlClassifier,
    SklearnToTsmlClusterer,
    SklearnToTsmlRegressor,
)


@parametrize_with_checks(
    [
        SklearnToTsmlClassifier(classifier=RandomForestClassifier(n_estimators=5)),
        SklearnToTsmlRegressor(regressor=RandomForestRegressor(n_estimators=5)),
        SklearnToTsmlClusterer(clusterer=KMeans(n_clusters=2, max_iter=5)),
    ]
)
def test_tsml_wrapper_estimator(estimator, check):
    """Test that tsml wrapper estimators adhere to tsml conventions."""
    check(estimator)



"""Collating results example."""

from tsml_eval.evaluation.multiple_estimator_evaluation import \
     evaluate_clusterers_by_problem, evaluate_classifiers_by_problem, evaluate_regressors_by_problem
from aeon.datasets.tsc_data_lists import univariate_equal_length

local_results = "C:\\ResultsWorkingArea\\"
ada_results = "/gpfs/home/ajb/ResultsWorkingArea/STC_compare/STC_COMPARE/"
ada_names = ["MAIN", "STC", "REF"]
local_names =["kmeans-erp","kmedoids-edr","kmedoids-erp"]
resamples = 30


def classifier_results(results, names, resamples=30, problems=univariate_equal_length):
    evaluate_classifiers_by_problem(
        load_path=results,
        classifier_names=names,
        dataset_names=list(problems),
        resamples=resamples,
        eval_name="cls_eval",
        save_path=results,
        verify_results=False,
        error_on_missing=False
    )

def regressor_results(results, names, resamples=30, problems=univariate_equal_length):
    evaluate_regressors_by_problem(
        load_path=results,
        regressor_names=names,
        dataset_names=list(univariate_equal_length),
        resamples=resamples,
        eval_name="reg_eval",
        save_path=results,
        verify_results=False,
        error_on_missing=False
    )


def clustering_results(results, names, resamples=10, problems=univariate_equal_length):
    evaluate_clusterers_by_problem(
        load_path=results,
        clusterer_names=names,
        dataset_names=list(univariate_equal_length),
        resamples=resamples,
        eval_name="reg_eval",
        save_path=results,
        verify_results=False,
        error_on_missing=False
    )


def extract_reference_results():
     classifier_names=names



"""Regressors used in the publication."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst", "dguijo"]

import numpy as np

from tsml_eval.utils.functions import str_in_nested_list

expansion_regressors = [
    ["1nn-dtw", "KNeighborsTimeSeriesRegressor"],
    "1nn-ed",
    "5nn-dtw",
    "5nn-ed",
    ["fcnn", "fcn", "fcnnregressor", "FCNRegressor"],
    ["FPCARegressor", "fpcregressor", "fpcr"],
    ["fpcar-b-spline", "fpcr-b-spline", "fpcr-bs"],
    ["grid-svr", "grid-supportvectorregressor"],
    [
        "inception",
        "singleinception",
        "IndividualInceptionRegressor",
        "individualinceptiontime",
    ],
    ["inceptione", "inception-e", "inceptiontime", "InceptionTimeRegressor"],
    ["rf", "randf", "randomforest", "RandomForestRegressor"],
    ["resnet", "ResNetRegressor"],
    ["rocket", "RocketRegressor"],
    ["multirocket", "multirocketregressor"],
    ["xgb", "xgboost", "xgboostregressor", "XGBRegressor"],
    ["cnn", "CNNRegressor"],
    ["RidgeCV", "ridge"],
    ["RotationForestRegressor", "rotf", "rotationforest"],
    ["tsf", "timeseriesforestregressor"],
    ["drcif", "DrCIFRegressor"],
    ["fresh-prince", "freshprince", "FreshPRINCERegressor"],
]


def _set_tser_exp_regressor(
    regressor_name,
    random_state=None,
    n_jobs=1,
    **kwargs,
):
    r = regressor_name.lower()

    if not str_in_nested_list(expansion_regressors, r):
        raise ValueError(f"UNKNOWN REGRESSOR: {r} in _set_tser_exp_regressor")

    if r == "1nn-dtw" or r == "kneighborstimeseriesregressor":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            n_neighbors=1,
            distance="dtw",
            distance_params={"window": 0.1},
            **kwargs,
        )
    elif r == "1nn-ed":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            distance="euclidean",
            n_neighbors=1,
            **kwargs,
        )
    elif r == "5nn-dtw":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            n_neighbors=5,
            distance="dtw",
            distance_params={"window": 0.1},
            **kwargs,
        )
    elif r == "5nn-ed":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            distance="euclidean",
            n_neighbors=5,
            **kwargs,
        )
    elif r == "fcnn" or r == "fcn" or r == "fcnnregressor" or r == "fcnregressor":
        from aeon.regression.deep_learning import FCNRegressor

        return FCNRegressor(
            random_state=random_state,
            **kwargs,
        )
    elif r == "fpcaregressor" or r == "fpcregressor" or r == "fpcr":
        from tsml.feature_based import FPCARegressor

        return FPCARegressor(
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "fpcar-b-spline" or r == "fpcr-b-spline" or r == "fpcr-bs":
        from tsml.feature_based import FPCARegressor

        return FPCARegressor(
            n_jobs=n_jobs,
            bspline=True,
            order=4,
            n_basis=10,
            **kwargs,
        )
    elif r == "grid-svr" or r == "grid-supportvectorregressor":
        from sklearn.model_selection import GridSearchCV
        from sklearn.svm import SVR

        param_grid = [
            {
                "kernel": ["rbf", "sigmoid"],
                "C": [0.1, 1, 10, 100],
                "gamma": [0.001, 0.01, 0.1, 1],
            }
        ]

        return GridSearchCV(
            SVR(
                **kwargs,
            ),
            param_grid,
            scoring="neg_mean_squared_error",
            cv=3,
            n_jobs=n_jobs,
        )
    elif (
        r == "inception"
        or r == "singleinception"
        or r == "individualinceptionregressor"
        or r == "individualinceptiontime"
    ):
        from aeon.regression.deep_learning import IndividualInceptionRegressor

        return IndividualInceptionRegressor(
            random_state=random_state,
            **kwargs,
        )
    elif (
        r == "inceptione"
        or r == "inception-e"
        or r == "inceptiontime"
        or r == "inceptiontimeregressor"
    ):
        from aeon.regression.deep_learning import InceptionTimeRegressor

        return InceptionTimeRegressor(
            random_state=random_state,
            **kwargs,
        )
    elif (
        r == "rf" or r == "randf" or r == "randomforest" or r == "randomforestregressor"
    ):
        from sklearn.ensemble import RandomForestRegressor

        return RandomForestRegressor(
            n_estimators=500,
            n_jobs=n_jobs,
            random_state=random_state,
            **kwargs,
        )
    elif r == "resnet" or r == "resnetregressor":
        from aeon.regression.deep_learning import ResNetRegressor

        return ResNetRegressor(
            random_state=random_state,
            **kwargs,
        )
    elif r == "rocket" or r == "rocketregressor":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "multirocket" or r == "multirocketregressor":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(
            rocket_transform="multirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "xgb" or r == "xgboost" or r == "xgboostregressor" or r == "xgbregressor":
        from xgboost import XGBRegressor

        return XGBRegressor(
            n_estimators=500,
            n_jobs=n_jobs,
            learning_rate=0.1,
            random_state=random_state,
            **kwargs,
        )
    elif r == "cnn" or r == "cnnregressor":
        from aeon.regression.deep_learning import CNNRegressor

        return CNNRegressor(
            random_state=random_state,
            **kwargs,
        )
    elif r == "ridgecv" or r == "ridge":
        from sklearn.linear_model import RidgeCV

        return RidgeCV(
            fit_intercept=True,
            alphas=np.logspace(-3, 3, 10),
            **kwargs,
        )
    elif r == "rotationforestregressor" or r == "rotf" or r == "rotationforest":
        from aeon.regression.sklearn import RotationForestRegressor

        return RotationForestRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "tsf" or r == "timeseriesforestregressor":
        from aeon.regression.interval_based import TimeSeriesForestRegressor
        from tsml.compose import ChannelEnsembleRegressor

        estimators = [
            (
                "tsf",
                TimeSeriesForestRegressor(
                    random_state=random_state,
                    n_estimators=500,
                    n_jobs=n_jobs,
                    **kwargs,
                ),
                None,
            )
        ]
        return ChannelEnsembleRegressor(estimators)
    elif r == "drcif" or r == "drcifregressor":
        from aeon.regression.interval_based import DrCIFRegressor

        return DrCIFRegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "fresh-prince" or r == "freshprince" or r == "freshprinceregressor":
        from aeon.regression.feature_based import FreshPRINCERegressor

        return FreshPRINCERegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )



"""Utilities for datasets."""

__author__ = ["MatthewMiddlehurst"]
__all__ = [
    "load_experiment_data",
    "copy_dataset_ts_files",
    "save_merged_dataset_splits",
]

import os
import shutil
from os.path import exists
from typing import List, Optional, Union

import numpy as np
from aeon.datasets import load_from_tsfile, write_to_tsfile


def load_experiment_data(
    problem_path: str,
    dataset: str,
    resample_id: int,
    predefined_resample: bool,
):
    """Load data for experiments.

    Parameters
    ----------
    problem_path : str
        Path to the problem folder.
    dataset : str
        Name of the dataset.
    resample_id : int or None
        Id of the data resample to use.
    predefined_resample : boolean
        If True, use the predefined resample.

    Returns
    -------
    X_train : np.ndarray or list of np.ndarray
        Train data in a 2d or 3d ndarray or list of arrays.
    y_train : np.ndarray
        Train data labels.
    X_test : np.ndarray or list of np.ndarray
        Test data in a 2d or 3d ndarray or list of arrays.
    y_test : np.ndarray
        Test data labels.
    resample : boolean
        If True, the data is to be resampled.
    """
    if resample_id is not None and predefined_resample:
        resample_str = "" if resample_id is None else str(resample_id)

        X_train, y_train = load_from_tsfile(
            f"{problem_path}/{dataset}/{dataset}{resample_str}_TRAIN.ts"
        )
        X_test, y_test = load_from_tsfile(
            f"{problem_path}/{dataset}/{dataset}{resample_str}_TEST.ts"
        )

        resample_data = False
    else:
        X_train, y_train = load_from_tsfile(
            f"{problem_path}/{dataset}/{dataset}_TRAIN.ts"
        )
        X_test, y_test = load_from_tsfile(f"{problem_path}/{dataset}/{dataset}_TEST.ts")

        resample_data = True if resample_id != 0 else False

    return X_train, y_train, X_test, y_test, resample_data


def copy_dataset_ts_files(
    datasets: Union[List[str], str],
    source_path: str,
    destination_path: str,
):
    """Copy the TRAIN and TEST .ts files of the datasets to the destination path.

    Expects files to be present at {path}/{dataset}/{dataset}_TRAIN.ts and
    {path}/{dataset}/{dataset}_TEST.ts.

    Parameters
    ----------
    datasets : list of str or str
        The names of the datasets to copy. If a list, each item is the name of a
        dataset. If a string, it is the path to a file containing the names of the
        datasets, one per line.
    source_path : str
        The path to the directory containing the datasets.
    destination_path : str
        The path to the directory where the datasets will be copied.

    Examples
    --------
    >>> from tsml_eval.testing.testing_utils import _TEST_DATA_PATH, _TEST_OUTPUT_PATH
    >>> from tsml_eval.utils.datasets import copy_dataset_ts_files
    >>> copy_dataset_ts_files(
    ...     f"{_TEST_DATA_PATH}/_test_data/test_datalist.txt",
    ...     _TEST_DATA_PATH,
    ...     f"{_TEST_OUTPUT_PATH}/datasets/",
    ... )
    """
    if isinstance(datasets, str):
        with open(datasets) as f:
            datasets = f.readlines()
            datasets = [d.strip() for d in datasets]

    for file in datasets:
        data_file = f"{source_path}/{file}/{file}"
        os.makedirs(f"{destination_path}/{file}", exist_ok=True)

        if not exists(f"{data_file}_TRAIN.ts"):
            raise FileNotFoundError(f"File not found: {data_file}_TRAIN.ts")
        shutil.copy(
            f"{data_file}_TRAIN.ts", f"{destination_path}/{file}/{file}_TRAIN.ts"
        )

        if not exists(f"{data_file}_TEST.ts"):
            raise FileNotFoundError(f"File not found: {data_file}_TEST.ts")
        shutil.copy(f"{data_file}_TEST.ts", f"{destination_path}/{file}/{file}_TEST.ts")


def save_merged_dataset_splits(
    problem_path: str,
    dataset: str,
    save_path: Optional[str] = None,
):
    """Merge the TRAIN and TEST .ts files of a dataset and save the merged file.

    Expects files to be present at {path}/{dataset}/{dataset}_TRAIN.ts and
    {path}/{dataset}/{dataset}_TEST.ts.

    Parameters
    ----------
    problem_path : str
        Path to the problem folder.
    dataset : str
        Name of the dataset.
    save_path : str, default=None
        Path to save the merged dataset to. If None, the merged dataset will be saved
        in the same folder as the original datasets.
    """
    if save_path is None:
        save_path = problem_path

    X_train, y_train = load_from_tsfile(f"{problem_path}/{dataset}/{dataset}_TRAIN.ts")
    X_test, y_test = load_from_tsfile(f"{problem_path}/{dataset}/{dataset}_TEST.ts")

    os.makedirs(save_path, exist_ok=True)
    X = np.concatenate([X_train, X_test], axis=0)
    y = np.concatenate([y_train, y_test], axis=0)

    write_to_tsfile(X, f"{save_path}/{dataset}/", y=y, problem_name=dataset)



"""Tests for validation utilities."""

from sklearn.cluster import KMeans
from sklearn.dummy import DummyClassifier, DummyRegressor
from tsml.interval_based import TSFClassifier

from tsml_eval.utils.estimator_validation import (
    is_sklearn_classifier,
    is_sklearn_clusterer,
    is_sklearn_estimator,
    is_sklearn_regressor,
)


def test_is_sklearn_estimator():
    """Test is_sklearn_estimator."""
    assert is_sklearn_estimator(DummyClassifier())
    assert not is_sklearn_estimator(TSFClassifier())


def test_is_sklearn_classifier():
    """Test is_sklearn_classifier."""
    assert is_sklearn_classifier(DummyClassifier())
    assert not is_sklearn_classifier(DummyRegressor())


def test_is_sklearn_regressor():
    """Test is_sklearn_regressor."""
    assert is_sklearn_regressor(DummyRegressor())
    assert not is_sklearn_regressor(DummyClassifier())


def test_is_sklearn_clusterer():
    """Test is_sklearn_clusterer."""
    assert is_sklearn_clusterer(KMeans())
    assert not is_sklearn_clusterer(DummyClassifier())



"""CICD numba caching functions."""

import pickle
import subprocess

import numba.core.caching


def get_invalid_numba_files():
    """Get the files that have been changed since the last commit.

    This function is used to get the files that have been changed. This is needed
    because custom logic to save the numba cache has been implemented for numba.
    This function returns the file names that have been changed and if they appear
    in here any numba functions cache are invalidated.

    Returns
    -------
    list
        List of file names that have been changed.
    """
    subprocess.run(["git", "fetch", "origin", "main"], check=True)

    result = subprocess.run(
        ["git", "diff", "--name-only", "origin/main"],
        check=True,
        capture_output=True,
        text=True,  # Makes the output text instead of bytes
    )

    files = result.stdout.split("\n")

    files = [file for file in files if file]

    clean_files = []

    for temp in files:
        if temp.endswith(".py"):
            clean_files.append((temp.split("/")[-1]).strip(".py"))

    return clean_files


# Retry the git fetch and git diff commands in case of failure
retry = 0
while retry < 3:
    try:
        CHANGED_FILES = get_invalid_numba_files()
        break
    except subprocess.CalledProcessError:
        retry += 1

# If the retry count is reached, raise an error
if retry == 3:
    raise Exception("Failed to get the changed files from git.")


def custom_load_index(self):
    """Overwrite load index method for numba.

    This is used to overwrite the numba internal logic to allow for caching during
    the cicd run. Numba traditionally checks the timestamp of the file and if it
    has changed it invalidates the cache. This is not ideal for the cicd run as
    the cache restore is always before the files (since it is cloned in) and
    thus the cache is always invalidated. This custom method ignores the timestamp
    element and instead just checks the file name. This isn't as fine grained as numba
    but it is better to invalidate more and make sure the right function has been
    compiled than try to be too clever and miss some.

    Returns
    -------
    dict
        Dictionary of the cached functions.
    """
    try:
        with open(self._index_path, "rb") as f:
            version = pickle.load(f)
            data = f.read()
    except FileNotFoundError:
        return {}
    if version != self._version:
        return {}
    stamp, overloads = pickle.loads(data)
    cache_filename = self._index_path.split("/")[-1].split("-")[0].split(".")[0]
    if stamp[1] != self._source_stamp[1] or cache_filename in CHANGED_FILES:
        return {}
    else:
        return overloads


original_load_index = numba.core.caching.IndexDataCacheFile._load_index
numba.core.caching.IndexDataCacheFile._load_index = custom_load_index


# Force all numba functions to be cached
original_jit = numba.core.decorators._jit


def custom_njit(*args, **kwargs):
    """Force jit to cache.

    This is used for libraries like stumpy that doesn't cache by default. This
    function will force all functions running to be cache'd
    """
    target = kwargs["targetoptions"]
    # This target can't be cached
    if "no_cpython_wrapper" not in target:
        kwargs["cache"] = True
    return original_jit(*args, **kwargs)


# Overwrite the jit function with the custom version
numba.core.decorators._jit = custom_njit



"""Set classifier function."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

from tsml_eval.utils.functions import str_in_nested_list

convolution_based_classifiers = [
    ["rocketclassifier", "rocket"],
    ["minirocket", "mini-rocket"],
    ["multirocket", "multi-rocket"],
    ["arsenalclassifier", "arsenal"],
    ["miniarsenal", "mini-arsenal"],
    ["multiarsenal", "multi-arsenal"],
    ["hydraclassifier", "hydra"],
    ["multirockethydraclassifier", "multirockethydra", "multirocket-hydra"],
]
deep_learning_classifiers = [
    ["cnnclassifier", "cnn"],
    ["fcnclassifier", "fcnn"],
    ["mlpclassifier", "mlp"],
    ["encoderclassifier", "encoder"],
    ["tapnetclassifier", "tapnet"],
    ["resnetclassifier", "resnet"],
    ["individualinceptionclassifier", "singleinception"],
    ["inceptiontimeclassifier", "inceptiontime"],
    ["h-inceptiontimeclassifier", "h-inceptiontime"],
    ["litetimeclassifier", "litetime"],
]
dictionary_based_classifiers = [
    ["bossensemble", "boss"],
    "individualboss",
    ["contractableboss", "cboss"],
    ["temporaldictionaryensemble", "tde"],
    "individualtde",
    "weasel",
    "weasel-logistic",
    "muse",
    "muse-logistic",
    ["weasel_v2", "weaseldilation", "weasel-dilation", "weasel-d"],
    "redcomets",
    "redcomets-500",
]
distance_based_classifiers = [
    ["kneighborstimeseriesclassifier", "dtw", "1nn-dtw"],
    ["ed", "1nn-euclidean", "1nn-ed"],
    ["msm", "1nn-msm"],
    ["twe", "1nn-twe"],
    "1nn-dtw-cv",
    ["elasticensemble", "ee"],
    ["grailclassifier", "grail"],
]
feature_based_classifiers = [
    "summary-500",
    ["summaryclassifier", "summary"],
    "catch22-500",
    ["catch22classifier", "catch22"],
    ["freshprinceclassifier", "freshprince"],
    "freshprince-500",
    "tsfresh-nofs",
    ["tsfreshclassifier", "tsfresh"],
    ["signatureclassifier", "signatures"],
]
hybrid_classifiers = [
    ["hivecotev1", "hc1"],
    ["hivecotev2", "hc2"],
    ["ristclassifier", "rist", "rist-extrat"],
]
interval_based_classifiers = [
    "rstsf-500",
    ["rstsfclassifier", "rstsf", "r-stsf"],
    "rise-500",
    ["randomintervalspectralensembleclassifier", "rise"],
    "tsf-500",
    ["timeseriesforestclassifier", "tsf"],
    "cif-500",
    ["canonicalintervalforestclassifier", "cif"],
    "stsf-500",
    ["supervisedtimeseriesforest", "stsf"],
    "drcif-500",
    ["drcif", "drcifclassifier"],
    "summary-intervals",
    ["randomintervals-500", "catch22-intervals-500"],
    ["randomintervalclassifier", "randomintervals", "catch22-intervals"],
    ["quantclassifier", "quant"],
]
other_classifiers = [
    ["dummyclassifier", "dummy", "dummyclassifier-aeon"],
    "dummyclassifier-tsml",
    "dummyclassifier-sklearn",
]
shapelet_based_classifiers = [
    "stc-2hour",
    ["shapelettransformclassifier", "stc"],
    ["rdstclassifier", "rdst"],
    ["randomshapeletforestclassifier", "randomshapeletforest", "rsf"],
    ["mrsqmclassifier", "mrsqm"],
    ["sastclassifier", "sast"],
    ["rsastclassifier", "rsast"],
]
vector_classifiers = [
    ["rotationforestclassifier", "rotationforest", "rotf"],
    ["ridgeclassifiercv", "ridgecv"],
    ["logisticregression", "logistic"],
]


def get_classifier_by_name(
    classifier_name,
    random_state=None,
    n_jobs=1,
    fit_contract=0,
    checkpoint=None,
    **kwargs,
):
    """Return a classifier matching a given input name.

    Basic way of creating a classifier to build using the default or alternative
    settings. This set up is to help with batch jobs for multiple problems and to
    facilitate easy reproducibility for use with run_classification_experiment.

    Generally, inputting a classifier class name will return said classifier with
    default settings.

    Parameters
    ----------
    classifier_name : str
        String indicating which classifier to be returned.
    random_state : int, RandomState instance or None, default=None
        Random seed or RandomState object to be used in the classifier if available.
    n_jobs: int, default=1
        The number of jobs to run in parallel for both classifier ``fit`` and
        ``predict`` if available. `-1` means using all processors.
    fit_contract: int, default=0
        Contract time in minutes for classifier ``fit`` if available.
    checkpoint: str or None, default=None
        Path to a checkpoint file to save the classifier if available. No checkpointing
        if None.
    **kwargs
        Additional keyword arguments to be passed to the classifier.

    Return
    ------
    classifier : A BaseClassifier.
        The classifier matching the input classifier name.
    """
    c = classifier_name.casefold()

    if str_in_nested_list(convolution_based_classifiers, c):
        return _set_classifier_convolution_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(deep_learning_classifiers, c):
        return _set_classifier_deep_learning(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(dictionary_based_classifiers, c):
        return _set_classifier_dictionary_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(distance_based_classifiers, c):
        return _set_classifier_distance_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(feature_based_classifiers, c):
        return _set_classifier_feature_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(hybrid_classifiers, c):
        return _set_classifier_hybrid(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(interval_based_classifiers, c):
        return _set_classifier_interval_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(other_classifiers, c):
        return _set_classifier_other(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(shapelet_based_classifiers, c):
        return _set_classifier_shapelet_based(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(vector_classifiers, c):
        return _set_classifier_vector(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    else:
        raise ValueError(f"UNKNOWN CLASSIFIER: {c} in set_classifier")


def _set_classifier_convolution_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "rocketclassifier" or c == "rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "minirocket" or c == "mini-rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(
            rocket_transform="minirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "multirocket" or c == "multi-rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(
            rocket_transform="multirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "arsenalclassifier" or c == "arsenal":
        from aeon.classification.convolution_based import Arsenal

        return Arsenal(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "miniarsenal" or c == "mini-arsenal":
        from aeon.classification.convolution_based import Arsenal

        return Arsenal(
            rocket_transform="minirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "multiarsenal" or c == "multi-arsenal":
        from aeon.classification.convolution_based import Arsenal

        return Arsenal(
            rocket_transform="multirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "hydraclassifier" or c == "hydra":
        from aeon.classification.convolution_based import HydraClassifier

        return HydraClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif (
        c == "multirockethydraclassifier"
        or c == "multirockethydra"
        or c == "multirocket-hydra"
    ):
        from aeon.classification.convolution_based import MultiRocketHydraClassifier

        return MultiRocketHydraClassifier(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )


def _set_classifier_deep_learning(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "cnnclassifier" or c == "cnn":
        from aeon.classification.deep_learning import CNNClassifier

        return CNNClassifier(random_state=random_state, **kwargs)
    elif c == "fcnclassifier" or c == "fcnn":
        from aeon.classification.deep_learning import FCNClassifier

        return FCNClassifier(random_state=random_state, **kwargs)
    elif c == "mlpclassifier" or c == "mlp":
        from aeon.classification.deep_learning import MLPClassifier

        return MLPClassifier(random_state=random_state, **kwargs)
    elif c == "encoderclassifier" or c == "encoder":
        from aeon.classification.deep_learning import EncoderClassifier

        return EncoderClassifier(random_state=random_state, **kwargs)
    elif c == "tapnetclassifier" or c == "tapnet":
        from aeon.classification.deep_learning import TapNetClassifier

        return TapNetClassifier(random_state=random_state, **kwargs)
    elif c == "resnetclassifier" or c == "resnet":
        from aeon.classification.deep_learning import ResNetClassifier

        return ResNetClassifier(random_state=random_state, **kwargs)
    elif c == "individualinceptionclassifier" or c == "singleinception":
        from aeon.classification.deep_learning import IndividualInceptionClassifier

        return IndividualInceptionClassifier(random_state=random_state, **kwargs)
    elif c == "inceptiontimeclassifier" or c == "inceptiontime":
        from aeon.classification.deep_learning import InceptionTimeClassifier

        return InceptionTimeClassifier(random_state=random_state, **kwargs)
    elif c == "h-inceptiontimeclassifier" or c == "h-inceptiontime":
        from aeon.classification.deep_learning import InceptionTimeClassifier

        return InceptionTimeClassifier(
            use_custom_filters=True, random_state=random_state, **kwargs
        )
    elif c == "litetimeclassifier" or c == "litetime":
        from aeon.classification.deep_learning import LITETimeClassifier

        return LITETimeClassifier(random_state=random_state, **kwargs)


def _set_classifier_dictionary_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "bossensemble" or c == "boss":
        from aeon.classification.dictionary_based import BOSSEnsemble

        return BOSSEnsemble(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "individualboss":
        from aeon.classification.dictionary_based import IndividualBOSS

        return IndividualBOSS(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "contractableboss" or c == "cboss":
        from aeon.classification.dictionary_based import ContractableBOSS

        return ContractableBOSS(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "temporaldictionaryensemble" or c == "tde":
        from aeon.classification.dictionary_based import TemporalDictionaryEnsemble

        return TemporalDictionaryEnsemble(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "individualtde":
        from aeon.classification.dictionary_based import IndividualTDE

        return IndividualTDE(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "weasel":
        from aeon.classification.dictionary_based import WEASEL

        return WEASEL(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "weasel-logistic":
        from aeon.classification.dictionary_based import WEASEL

        return WEASEL(
            random_state=random_state,
            n_jobs=n_jobs,
            support_probabilities=True,
            **kwargs,
        )
    elif c == "muse":
        from aeon.classification.dictionary_based import MUSE

        return MUSE(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "muse-logistic":
        from aeon.classification.dictionary_based import MUSE

        return MUSE(
            random_state=random_state,
            n_jobs=n_jobs,
            support_probabilities=True,
            **kwargs,
        )
    elif (
        c == "weasel_v2"
        or c == "weaseldilation"
        or c == "weasel-dilation"
        or c == "weasel-d"
    ):
        from aeon.classification.dictionary_based import WEASEL_V2

        return WEASEL_V2(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "redcomets":
        from aeon.classification.dictionary_based import REDCOMETS

        return REDCOMETS(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "redcomets-500":
        from aeon.classification.dictionary_based import REDCOMETS

        return REDCOMETS(
            n_trees=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )


def _set_classifier_distance_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "kneighborstimeseriesclassifier" or c == "dtw" or c == "1nn-dtw":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier

        return KNeighborsTimeSeriesClassifier(distance="dtw", n_jobs=n_jobs, **kwargs)
    elif c == "ed" or c == "1nn-euclidean" or c == "1nn-ed":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier

        return KNeighborsTimeSeriesClassifier(
            distance="euclidean", n_jobs=n_jobs, **kwargs
        )
    elif c == "msm" or c == "1nn-msm":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier

        return KNeighborsTimeSeriesClassifier(distance="msm", n_jobs=n_jobs, **kwargs)
    elif c == "twe" or c == "1nn-twe":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier

        return KNeighborsTimeSeriesClassifier(distance="twe", n_jobs=n_jobs, **kwargs)
    elif c == "elasticensemble" or c == "ee":
        from aeon.classification.distance_based import ElasticEnsemble

        return ElasticEnsemble(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "1nn-dtw-cv":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier
        from sklearn.model_selection import GridSearchCV

        param_grid = {"distance_params": [{"window": x / 100} for x in range(0, 100)]}
        return GridSearchCV(
            estimator=KNeighborsTimeSeriesClassifier(),
            param_grid=param_grid,
            scoring="accuracy",
            **kwargs,
        )
    elif c == "grailclassifier" or c == "grail":
        from tsml.distance_based import GRAILClassifier

        return GRAILClassifier(**kwargs)


def _set_classifier_feature_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "summary-500":
        from aeon.classification.feature_based import SummaryClassifier
        from sklearn.ensemble import RandomForestClassifier

        return SummaryClassifier(
            estimator=RandomForestClassifier(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "summaryclassifier" or c == "summary":
        from aeon.classification.feature_based import SummaryClassifier

        return SummaryClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "catch22-500":
        from aeon.classification.feature_based import Catch22Classifier
        from sklearn.ensemble import RandomForestClassifier

        return Catch22Classifier(
            estimator=RandomForestClassifier(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "catch22classifier" or c == "catch22":
        from aeon.classification.feature_based import Catch22Classifier

        return Catch22Classifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "freshprinceclassifier" or c == "freshprince":
        from aeon.classification.feature_based import FreshPRINCEClassifier

        return FreshPRINCEClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "freshprince-500":
        from aeon.classification.feature_based import FreshPRINCEClassifier

        return FreshPRINCEClassifier(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif c == "tsfresh-nofs":
        from aeon.classification.feature_based import TSFreshClassifier

        return TSFreshClassifier(
            relevant_feature_extractor=False,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "tsfreshclassifier" or c == "tsfresh":
        from aeon.classification.feature_based import TSFreshClassifier

        return TSFreshClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "signatureclassifier" or c == "signatures":
        from aeon.classification.feature_based import SignatureClassifier

        return SignatureClassifier(random_state=random_state, **kwargs)


def _set_classifier_hybrid(c, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if c == "hivecotev1" or c == "hc1":
        from aeon.classification.hybrid import HIVECOTEV1

        return HIVECOTEV1(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "hivecotev2" or c == "hc2":
        from aeon.classification.hybrid import HIVECOTEV2

        return HIVECOTEV2(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "ristclassifier" or c == "rist" or c == "rist-extrat":
        from aeon.classification.hybrid import RISTClassifier
        from sklearn.ensemble import ExtraTreesClassifier

        return RISTClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=ExtraTreesClassifier(n_estimators=500, criterion="entropy"),
            **kwargs,
        )


def _set_classifier_interval_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "rstsf-500":
        from aeon.classification.interval_based import RSTSF

        return RSTSF(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif c == "rstsfclassifier" or c == "rstsf" or c == "r-stsf":
        from aeon.classification.interval_based import RSTSF

        return RSTSF(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "rise-500":
        from aeon.classification.interval_based import (
            RandomIntervalSpectralEnsembleClassifier,
        )

        return RandomIntervalSpectralEnsembleClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "randomintervalspectralensembleclassifier" or c == "rise":
        from aeon.classification.interval_based import (
            RandomIntervalSpectralEnsembleClassifier,
        )

        return RandomIntervalSpectralEnsembleClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "tsf-500":
        from aeon.classification.interval_based import TimeSeriesForestClassifier

        return TimeSeriesForestClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "timeseriesforestclassifier" or c == "tsf":
        from aeon.classification.interval_based import TimeSeriesForestClassifier

        return TimeSeriesForestClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "cif-500":
        from aeon.classification.interval_based import CanonicalIntervalForestClassifier

        return CanonicalIntervalForestClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "canonicalintervalforestclassifier" or c == "cif":
        from aeon.classification.interval_based import CanonicalIntervalForestClassifier

        return CanonicalIntervalForestClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "stsf-500":
        from aeon.classification.interval_based import SupervisedTimeSeriesForest

        return SupervisedTimeSeriesForest(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "supervisedtimeseriesforest" or c == "stsf":
        from aeon.classification.interval_based import SupervisedTimeSeriesForest

        return SupervisedTimeSeriesForest(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "drcif-500":
        from aeon.classification.interval_based import DrCIFClassifier

        return DrCIFClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "drcif" or c == "drcifclassifier":
        from aeon.classification.interval_based import DrCIFClassifier

        return DrCIFClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "summary-intervals":
        from aeon.classification.interval_based import RandomIntervalClassifier
        from aeon.transformations.collection.feature_based import (
            SevenNumberSummaryTransformer,
        )
        from sklearn.ensemble import RandomForestClassifier

        return RandomIntervalClassifier(
            features=SevenNumberSummaryTransformer(),
            estimator=RandomForestClassifier(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "randomintervals-500" or c == "catch22-intervals-500":
        from aeon.classification.interval_based import RandomIntervalClassifier
        from sklearn.ensemble import RandomForestClassifier

        return RandomIntervalClassifier(
            estimator=RandomForestClassifier(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif (
        c == "randomintervalclassifier"
        or c == "randomintervals"
        or c == "catch22-intervals"
    ):
        from aeon.classification.interval_based import RandomIntervalClassifier

        return RandomIntervalClassifier(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif c == "quantclassifier" or c == "quant":
        from aeon.classification.interval_based import QUANTClassifier

        return QUANTClassifier(random_state=random_state, **kwargs)


def _set_classifier_other(c, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if c == "dummyclassifier" or c == "dummy" or c == "dummyclassifier-aeon":
        from aeon.classification import DummyClassifier

        return DummyClassifier(random_state=random_state, **kwargs)
    elif c == "dummyclassifier-tsml":
        from tsml.dummy import DummyClassifier

        return DummyClassifier(random_state=random_state, **kwargs)
    elif c == "dummyclassifier-sklearn":
        from sklearn.dummy import DummyClassifier

        return DummyClassifier(random_state=random_state, **kwargs)


def _set_classifier_shapelet_based(
    c, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if c == "stc-2hour":
        from aeon.classification.shapelet_based import ShapeletTransformClassifier

        return ShapeletTransformClassifier(
            transform_limit_in_minutes=120,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "shapelettransformclassifier" or c == "stc":
        from aeon.classification.shapelet_based import ShapeletTransformClassifier

        return ShapeletTransformClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "rdstclassifier" or c == "rdst":
        from aeon.classification.shapelet_based import RDSTClassifier

        return RDSTClassifier(random_state=random_state, **kwargs)
    elif (
        c == "randomshapeletforestclassifier"
        or c == "randomshapeletforest"
        or c == "rsf"
    ):
        from tsml.shapelet_based import RandomShapeletForestClassifier

        return RandomShapeletForestClassifier(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif c == "mrsqmclassifier" or c == "mrsqm":
        from aeon.classification.shapelet_based import MrSQMClassifier

        return MrSQMClassifier(random_state=random_state, **kwargs)
    elif c == "sastclassifier" or c == "sast":
        from aeon.classification.shapelet_based import SASTClassifier

        return SASTClassifier(seed=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "rsastclassifier" or c == "rsast":
        from aeon.classification.shapelet_based import RSASTClassifier

        return RSASTClassifier(seed=random_state, n_jobs=n_jobs, **kwargs)


def _set_classifier_vector(c, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if c == "rotationforestclassifier" or c == "rotationforest" or c == "rotf":
        from tsml.vector import RotationForestClassifier

        return RotationForestClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif c == "ridgeclassifiercv" or c == "ridgecv":
        from sklearn.linear_model import RidgeClassifierCV

        return RidgeClassifierCV(**kwargs)
    elif c == "logisticregression" or c == "logistic":
        from sklearn.linear_model import LogisticRegression

        return LogisticRegression(random_state=random_state, n_jobs=n_jobs, **kwargs)



"""Set forecaster function."""

__author__ = ["MatthewMiddlehurst"]

from aeon.forecasting.naive import NaiveForecaster

from tsml_eval.utils.functions import str_in_nested_list

ml_forecasters = [
    ["lr", "linearregression"],
    "1nn",
    ["rf", "randomforest"],
    "xgboost",
    "inceptiontime",
    "rocket",
    "freshprince",
    "drcif",
    ["rotf", "rotationforest"],
]
other_forecasters = [
    ["naiveforecaster", "naive"],
]


def get_forecaster_by_name(forecaster_name, random_state=None, n_jobs=1, **kwargs):
    """Return a forecaster matching a given input name.

    Basic way of creating a forecaster to build using the default or alternative
    settings. This set up is to help with batch jobs for multiple problems and to
    facilitate easy reproducibility for use with run_forecasting_experiment.

    Generally, inputting a forecasters class name will return said forecaster with
    default settings.

    todo

    Parameters
    ----------
    forecaster_name : str
        String indicating which classifier to be returned.
    random_state : int, RandomState instance or None, default=None
        Random seed or RandomState object to be used in the classifier if available.
    n_jobs: int, default=1
        The number of jobs to run in parallel for both classifier ``fit`` and
        ``predict`` if available. `-1` means using all processors.

    Return
    ------
    forecaster : A BaseForecaster.
        The forecaster matching the input classifier name.
    """
    f = forecaster_name.lower()

    if str_in_nested_list(ml_forecasters, f):
        return _set_forecaster_ml(f, random_state, n_jobs, kwargs)
    elif str_in_nested_list(other_forecasters, f):
        return _set_forecaster_other(f, random_state, n_jobs, kwargs)
    else:
        raise ValueError(f"UNKNOWN FORECASTER: {f} in set_forecaster")


def _set_forecaster_ml(f, random_state, n_jobs, kwargs):
    if f == "lr" or f == "linearregression":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.transformations.detrend import Detrender
        from sklearn.linear_model import LinearRegression
        from sklearn.preprocessing import StandardScaler

        regressor = LinearRegression(n_jobs=n_jobs, **kwargs)
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "1nn":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.transformations.detrend import Detrender
        from sklearn.neighbors import KNeighborsRegressor
        from sklearn.preprocessing import StandardScaler

        regressor = KNeighborsRegressor(n_neighbors=1, n_jobs=n_jobs, **kwargs)
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "rf" or f == "randomforest":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.transformations.detrend import Detrender
        from sklearn.ensemble import RandomForestRegressor
        from sklearn.preprocessing import StandardScaler

        regressor = RandomForestRegressor(
            n_estimators=200, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "xgboost":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler
        from xgboost import XGBRegressor

        regressor = XGBRegressor(
            n_estimators=200, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )

    elif f == "inceptiontime":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.regression.deep_learning import InceptionTimeRegressor
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler

        regressor = InceptionTimeRegressor(random_state=random_state, **kwargs)
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "rocket":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.regression.convolution_based import RocketRegressor
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler

        regressor = RocketRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "freshprince":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.regression.feature_based import FreshPRINCERegressor
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler

        regressor = FreshPRINCERegressor(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "drcif":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.regression.interval_based import DrCIFRegressor
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler

        regressor = DrCIFRegressor(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )
    elif f == "rotf" or f == "rotationforest":
        from aeon.forecasting.compose import make_reduction
        from aeon.forecasting.trend import PolynomialTrendForecaster
        from aeon.pipeline import make_pipeline
        from aeon.regression.sklearn import RotationForestRegressor
        from aeon.transformations.detrend import Detrender
        from sklearn.preprocessing import StandardScaler

        regressor = RotationForestRegressor(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )
        return make_pipeline(
            Detrender(PolynomialTrendForecaster(degree=1, with_intercept=True)),
            StandardScaler(),
            make_reduction(regressor, window_length=15, strategy="recursive"),
        )


def _set_forecaster_other(f, random_state, n_jobs, kwargs):
    if f == "naiveforecaster" or f == "naive":
        return NaiveForecaster(**kwargs)



"""General tests for the experiments module."""

import os

from tsml.dummy import DummyClassifier

from tsml_eval.experiments import (
    classification_experiments,
    load_and_run_classification_experiment,
)
from tsml_eval.experiments.tests import _CLASSIFIER_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH, _TEST_OUTPUT_PATH
from tsml_eval.utils.tests.test_results_writing import _check_classification_file_format


def test_kwargs():
    """Test experiments with kwargs input."""
    dataset = "MinimalChinatown"
    classifier = "LogisticRegression"

    result_path = _TEST_OUTPUT_PATH + "/kwargs/"

    args = [
        _TEST_DATA_PATH,
        result_path,
        classifier,
        dataset,
        "0",
        "--kwargs",
        "fit_intercept",
        "False",
        "bool",
        "--kwargs",
        "C",
        "0.8",
        "float",
        "--kwargs",
        "max_iter",
        "10",
        "int",
        "-ow",
    ]

    classification_experiments.run_experiment(args)

    test_file = f"{result_path}{classifier}/Predictions/{dataset}/testResample0.csv"

    assert os.path.exists(test_file)
    os.remove(test_file)


def test_experiments_predefined_resample_data_loading():
    """Test experiments with data loading."""
    dataset = "PredefinedChinatown"

    load_and_run_classification_experiment(
        _TEST_DATA_PATH + "_test_data/",
        _CLASSIFIER_RESULTS_PATH,
        dataset,
        DummyClassifier(),
        resample_id=5,
        predefined_resample=True,
    )

    test_file = (
        f"{_CLASSIFIER_RESULTS_PATH}/DummyClassifier/Predictions/{dataset}/"
        "testResample5.csv"
    )
    assert os.path.exists(test_file)
    _check_classification_file_format(test_file)

    os.remove(test_file)



"""Files for TSER expansion publication."""

__all__ = [
    "_set_tser_exp_regressor",
    "_run_experiment",
    "expansion_regressors",
]

from tsml_eval.publications.y2023.tser_archive_expansion.run_experiments import (
    _run_experiment,
)
from tsml_eval.publications.y2023.tser_archive_expansion.set_tser_exp_regressor import (
    _set_tser_exp_regressor,
    expansion_regressors,
)



"""Tests for arguments parsing."""

import pytest

from tsml_eval.testing.testing_utils import suppress_output
from tsml_eval.utils.arguments import parse_args


def test_positional_args():
    """Test parsing of positional arguments."""
    args = [
        "D:/data/",
        "D:/results/",
        "Est",
        "Dataset",
        "0",
    ]
    args = parse_args(args)

    assert args.data_path == "D:/data/"
    assert args.results_path == "D:/results/"
    assert args.estimator_name == "Est"
    assert args.dataset_name == "Dataset"
    assert args.resample_id == 0
    assert args.random_seed is None
    assert args.n_jobs == 1
    assert args.train_fold is False
    assert args.kwargs == {}


@pytest.mark.parametrize(
    "args",
    [
        ["D:/data/", "D:/results/", "Est", "Dataset"],
        ["D:/data/", "D:/results/", "Est", "Dataset", "zero"],
        ["D:/data/", "D:/results/", "Est", "Dataset", 1],
        ["D:/data/", "D:/results/", "Est", "Dataset", "0", "1"],
        [[1, 2], "D:/results/", "Est", "Dataset", "0"],
    ],
)
@suppress_output()
def test_wrong_positional_args(args):
    """Test parsing of incorrect positional arguments."""
    with pytest.raises((SystemExit, TypeError, ValueError)):
        parse_args(args)


def test_kw_args():
    """Test parsing of keyword arguments."""
    args = [
        "D:/data/",
        "D:/results/",
        "Est",
        "Dataset",
        "1",
        "-rs",
        "10",
        "--n_jobs",
        "4",
        "-tr",
        "--kwargs",
        "key1",
        "value1",
        "str",
        "-kw",
        "key2",
        "value2",
        "str",
    ]
    args = parse_args(args)

    assert args.data_path == "D:/data/"
    assert args.results_path == "D:/results/"
    assert args.estimator_name == "Est"
    assert args.dataset_name == "Dataset"
    assert args.resample_id == 1
    assert args.random_seed == 10
    assert args.n_jobs == 4
    assert args.train_fold is True
    assert args.kwargs["key1"] == "value1"
    assert args.kwargs["key2"] == "value2"


@pytest.mark.parametrize(
    "args",
    [
        ["D:/data/", "D:/results/", "Est", "Dataset", "1", "-rs"],
        ["D:/data/", "D:/results/", "Est", "Dataset", "1", "-rs", "zero"],
        ["D:/data/", "D:/results/", "Est", "Dataset", "1", "-tr", "True"],
        ["D:/data/", "D:/results/", "Est", "Dataset", "1", "-kw", "key value str"],
        ["D:/data/", "D:/results/", "Est", "Dataset", "1", "-kw", "key", "value"],
    ],
)
@suppress_output()
def test_wrong_kw_args(args):
    """Test parsing of incorrect keyword arguments."""
    with pytest.raises((SystemExit, TypeError, ValueError)):
        parse_args(args)



"""Set classifier function."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

import numpy as np
from aeon.clustering import (
    TimeSeriesCLARA,
    TimeSeriesCLARANS,
    TimeSeriesKMeans,
    TimeSeriesKMedoids,
)
from aeon.transformations.collection import TimeSeriesScaler
from sklearn.cluster import KMeans

from tsml_eval.utils.datasets import load_experiment_data
from tsml_eval.utils.functions import str_in_nested_list

distance_based_clusterers = [
    "kmeans-euclidean",
    "kmeans-squared",
    "kmeans-dtw",
    "kmeans-ddtw",
    "kmeans-wdtw",
    "kmeans-wddtw",
    "kmeans-lcss",
    "kmeans-erp",
    "kmeans-edr",
    "kmeans-twe",
    "kmeans-msm",
    "kmeans-adtw",
    "kmeans-shape_dtw",
    "kmedoids-euclidean",
    "kmedoids-squared",
    "kmedoids-dtw",
    "kmedoids-ddtw",
    "kmedoids-wdtw",
    "kmedoids-wddtw",
    "kmedoids-lcss",
    "kmedoids-erp",
    "kmedoids-edr",
    "kmedoids-twe",
    "kmedoids-msm",
    "kmedoids-adtw",
    "kmedoids-shape_dtw",
    "clarans-euclidean",
    "clarans-squared",
    "clarans-dtw",
    "clarans-ddtw",
    "clarans-wdtw",
    "clarans-wddtw",
    "clarans-lcss",
    "clarans-erp",
    "clarans-edr",
    "clarans-twe",
    "clarans-msm",
    "clarans-adtw",
    "clarans-shape_dtw",
    "clara-euclidean",
    "clara-squared",
    "clara-dtw",
    "clara-ddtw",
    "clara-wdtw",
    "clara-wddtw",
    "clara-lcss",
    "clara-erp",
    "clara-edr",
    "clara-twe",
    "clara-msm",
    "clara-adtw",
    "clara-shape_dtw",
    "pam-euclidean",
    "pam-squared",
    "pam-dtw",
    "pam-ddtw",
    "pam-wdtw",
    "pam-wddtw",
    "pam-lcss",
    "pam-erp",
    "pam-edr",
    "pam-twe",
    "pam-msm",
    "pam-adtw",
    "pam-shape_dtw",
    "kmeans-ba-dtw",
    "kmeans-ba-ddtw",
    "kmeans-ba-wdtw",
    "kmeans-ba-wddtw",
    "kmeans-ba-lcss",
    "kmeans-ba-erp",
    "kmeans-ba-edr",
    "kmeans-ba-twe",
    "kmeans-ba-msm",
    "kmeans-ba-adtw",
    "kmeans-ba-shape_dtw",
    "kmeans-ssg-ba-dtw",
    "kmeans-ssg-ba-ddtw",
    "kmeans-ssg-ba-wdtw",
    "kmeans-ssg-ba-wddtw",
    "kmeans-ssg-ba-erp",
    "kmeans-ssg-ba-edr",
    "kmeans-ssg-ba-twe",
    "kmeans-ssg-ba-msm",
    "kmeans-ssg-ba-adtw",
    "kmeans-ssg-ba-shape_dtw",
    "timeserieskmeans",
    "timeserieskmedoids",
    "timeseriesclarans",
    "timeseriesclara",
]

other_clusterers = [
    ["dummyclusterer", "dummy", "dummyclusterer-tsml"],
    "dummyclusterer-aeon",
    "dummyclusterer-sklearn",
]
vector_clusterers = [
    ["kmeans", "kmeans-sklearn"],
    "dbscan",
]


def get_clusterer_by_name(
    clusterer_name,
    random_state=None,
    n_jobs=1,
    fit_contract=0,
    checkpoint=None,
    data_vars=None,
    row_normalise=False,
    **kwargs,
):
    """Return a clusterer matching a given input name.

    Basic way of creating a clusterer to build using the default or alternative
    settings. This set up is to help with batch jobs for multiple problems and to
    facilitate easy reproducibility through run_clustering_experiment.

    Generally, inputting a clusterer class name will return said clusterer with
    default settings.

    Parameters
    ----------
    clusterer_name : str
        String indicating which clusterer to be returned.
    random_state : int, RandomState instance or None, default=None
        Random seed or RandomState object to be used in the clusterer if available.
    n_jobs: int, default=1
        The number of jobs to run in parallel for both clusterer ``fit`` and
        ``predict`` if available. `-1` means using all processors.
    fit_contract: int, default=0
        The number of data points to use in the clusterer ``fit`` if available.
    checkpoint: str, default=None
        Checkpoint to save model
    data_vars: list, default=None
        List of arguments to load the dataset using
        `tsml_eval.utils.experiments import load_experiment_data`.
    row_normalise: bool, default=False
        Whether to row normalise the data if it is loaded using data_vars.

    Return
    ------
    clusterer: A BaseClusterer.
        The clusterer matching the input clusterer name.
    """
    c = clusterer_name.lower()

    if str_in_nested_list(distance_based_clusterers, c):
        return _set_clusterer_distance_based(
            c,
            random_state,
            n_jobs,
            fit_contract,
            checkpoint,
            data_vars,
            row_normalise,
            kwargs,
        )
    elif str_in_nested_list(other_clusterers, c):
        return _set_clusterer_other(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(vector_clusterers, c):
        return _set_clusterer_vector(
            c, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    else:
        raise ValueError(f"UNKNOWN CLUSTERER: {c} in set_clusterer")


def _set_clusterer_distance_based(
    c,
    random_state,
    n_jobs,
    fit_contract,
    checkpoint,
    data_vars,
    row_normalise,
    kwargs,
):
    if "init_algorithm" in kwargs:
        init_algorithm = kwargs["init_algorithm"]
    else:
        init_algorithm = "random"

    if "distance" in kwargs:
        distance = kwargs["distance"]
    else:
        distance = c.split("-")[-1]

    if "distance_params" in kwargs:
        distance_params = kwargs["distance_params"]
    else:
        distance_params = _get_distance_default_params(
            distance, data_vars, row_normalise
        )

    if "kmeans" in c or "timeserieskmeans" in c:
        if "average_params" in kwargs:
            average_params = kwargs["average_params"]
        else:
            average_params = {"distance": distance, **distance_params.copy()}

        if "ssg" in c:
            # Sets to use subgradient BA
            average_params = {
                **average_params,
                "method": "subgradient",
            }
            return TimeSeriesKMeans(
                max_iter=50,
                n_init=10,
                init_algorithm=init_algorithm,
                distance=distance,
                distance_params=distance_params,
                random_state=random_state,
                averaging_method="ba",
                average_params=average_params,
                **kwargs,
            )
        elif "ba" in c:
            return TimeSeriesKMeans(
                max_iter=50,
                n_init=10,
                init_algorithm=init_algorithm,
                distance=distance,
                distance_params=distance_params,
                random_state=random_state,
                averaging_method="ba",
                average_params=average_params,
                **kwargs,
            )
        else:
            return TimeSeriesKMeans(
                max_iter=50,
                n_init=10,
                init_algorithm=init_algorithm,
                distance=distance,
                distance_params=distance_params,
                random_state=random_state,
                averaging_method="mean",
                **kwargs,
            )
    elif "kmedoids" in c or "timeserieskmedoids" in c:
        return TimeSeriesKMedoids(
            max_iter=50,
            n_init=10,
            init_algorithm=init_algorithm,
            distance=distance,
            distance_params=distance_params,
            random_state=random_state,
            method="alternate",
            **kwargs,
        )
    elif "pam" in c or "timeseriespam" in c:
        return TimeSeriesKMedoids(
            max_iter=50,
            n_init=10,
            init_algorithm=init_algorithm,
            distance=distance,
            distance_params=distance_params,
            random_state=random_state,
            method="pam",
            **kwargs,
        )
    elif "clarans" in c or "timeseriesclarans" in c:
        return TimeSeriesCLARANS(
            n_init=10,
            init_algorithm=init_algorithm,
            distance=distance,
            distance_params=distance_params,
            random_state=random_state,
            **kwargs,
        )
    elif "clara" in c or "timeseriesclara" in c:
        return TimeSeriesCLARA(
            max_iter=50,
            init_algorithm=init_algorithm,
            distance=distance,
            distance_params=distance_params,
            random_state=random_state,
            **kwargs,
        )
    return None


def _get_distance_default_params(
    dist_name: str, data_vars: list, row_normalise: bool
) -> dict:
    if dist_name == "dtw" or dist_name == "ddtw":
        return {"window": 0.2}
    if dist_name == "lcss":
        return {"epsilon": 1.0}
    if dist_name == "erp":
        # load dataset to get std if available
        if data_vars is not None:
            X_train, _, _, _, _ = load_experiment_data(*data_vars)

            # cant handle unequal length series
            if isinstance(X_train, np.ndarray):
                if row_normalise:
                    scaler = TimeSeriesScaler()
                    X_train = scaler.fit_transform(X_train)

                return {"g": X_train.std(axis=0).sum()}
            elif not isinstance(X_train, list):
                raise ValueError("Unknown data type in _get_distance_default_params")
        return {"g": 0.05}
    if dist_name == "msm":
        return {"c": 1.0, "independent": True}
    if dist_name == "edr":
        return {"epsilon": None}
    if dist_name == "twe":
        return {"nu": 0.001, "lmbda": 1.0}
    if dist_name == "psi_dtw":
        return {"r": 0.5}
    if dist_name == "adtw":
        return {"warp_penalty": 1.0}
    if dist_name == "shape_dtw":
        return {"descriptor": "identity", "reach": 30}
    return {}


def _set_clusterer_other(c, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if c == "dummyclusterer" or c == "dummy" or c == "dummyclusterer-tsml":
        from tsml.dummy import DummyClusterer

        return DummyClusterer(
            strategy="random", n_clusters=1, random_state=random_state, **kwargs
        )
    elif c == "dummyclusterer-aeon":
        return TimeSeriesKMeans(
            n_clusters=1,
            n_init=1,
            init_algorithm="random",
            distance="euclidean",
            max_iter=1,
            random_state=random_state,
            **kwargs,
        )
    elif c == "dummyclusterer-sklearn":
        return KMeans(
            n_clusters=1,
            n_init=1,
            init="random",
            max_iter=1,
            random_state=random_state,
            **kwargs,
        )


def _set_clusterer_vector(c, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if c == "kmeans" or c == "kmeans-sklearn":
        from sklearn.cluster import KMeans

        return KMeans(random_state=random_state, **kwargs)
    elif c == "dbscan":
        from sklearn.cluster import DBSCAN

        return DBSCAN(n_jobs=n_jobs, **kwargs)



"""IVC consensus clustering algorithm."""

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.cluster import KMeans
from sklearn.utils import check_random_state
from tsml.base import _clone_estimator


class IterativeVotingClustering(BaseEstimator, ClusterMixin):
    """
    IVC (Iterative Voting Clustering) Consensus Clusterer.

    IVC is a consensus clustering algorithm that combines the results of multiple
    base clusterers to find a consensus clustering. It iteratively refines cluster
    assignments based on a majority voting scheme.

    Parameters
    ----------
    clusterers : list of clusterers, default=None
        A list of clusterers to use in the ensemble. If None, defaults to 5
        KMeans clusterers.
    init : {'plus', 'random', 'aligned'}, default='plus'
        The method used to initialize the cluster centers. 'plus' uses the
        k-means++ initialization method, 'random' uses random selection, and
        'aligned' uses a method that aligns the cluster assignmeds from the base
        clusterers.
    n_clusters : int, default=8
        The number of clusters to form.
    max_iterations : int, default=500
        The maximum number of iterations to perform.
    random_state : int, default=None
        The seed for random number generation.

    Attributes
    ----------
    labels_ : ndarray of shape (n_instances,)
        Labels of each point from the last fit.

    Examples
    --------
    >>> from tsml_eval.estimators.clustering.consensus.ivc import (
    ...     IterativeVotingClustering
    ... )
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.metrics import rand_score
    >>> iris = load_iris()
    >>> ivc = IterativeVotingClustering(n_clusters=3, random_state=0)
    >>> ivc.fit(iris.data)
    IterativeVotingClustering(...)
    >>> rand_score(iris.target, ivc.labels_)
    0.8737360178970918
    """

    def __init__(
        self,
        clusterers=None,
        init="plus",
        n_clusters=8,
        max_iterations=500,
        random_state=None,
    ):
        self.clusterers = clusterers
        self.init = init
        self.n_clusters = n_clusters
        self.max_iterations = max_iterations
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit model to X using IVC."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "IterativeVotingClustering is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, ensure_min_samples=self.n_clusters)

        rng = check_random_state(self.random_state)

        if self.clusterers is None:
            self._clusterers = [
                KMeans(
                    n_clusters=self.n_clusters,
                    n_init="auto",
                    random_state=rng.randint(np.iinfo(np.int32).max),
                )
                for _ in range(5)
            ]
        else:
            self._clusterers = [
                _clone_estimator(
                    clusterer, random_state=rng.randint(np.iinfo(np.int32).max)
                )
                for clusterer in self.clusterers
            ]

        cluster_assignments = np.zeros(
            (len(self._clusterers), X.shape[0]), dtype=np.int32
        )
        for i, clusterer in enumerate(self._clusterers):
            clusterer.fit(X)
            cluster_assignments[i] = clusterer.labels_

            uc = np.unique(clusterer.labels_)
            if (np.sort(uc) != np.arange(self.n_clusters)).any():
                raise ValueError(
                    "Input clusterers must have cluster labels in the range "
                    "0 to  n_clusters - 1."
                )

        self._build_ensemble(cluster_assignments)

        return self

    def predict(self, X):
        """Predict cluster labels for X."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "IterativeVotingClustering is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, reset=False)

        cluster_assignments = np.zeros(
            (len(self._clusterers), X.shape[0]), dtype=np.int32
        )
        for i in range(0, len(self._clusterers)):
            cluster_assignments[i] = self._clusterers[i].predict(X)

        rng = check_random_state(self.random_state)
        labels = self._calculate_cluster_membership(cluster_assignments, rng)

        return labels

    def _build_ensemble(self, cluster_assignments):
        rng = check_random_state(self.random_state)

        if self.init == "plus":
            self._initial_cluster_centers_plus(cluster_assignments, rng)
        elif self.init == "random":
            self._initial_cluster_centers_random(cluster_assignments, rng)
        elif self.init == "aligned":
            self._initial_cluster_centers_aligned(cluster_assignments, rng)
        else:
            raise ValueError("Invalid init method")

        self.labels_ = np.full(cluster_assignments.shape[1], -1, dtype=np.int32)

        iterations = 0
        while iterations < self.max_iterations:
            if iterations > 0:
                self._select_cluster_centers(cluster_assignments, rng)

            labels = self._calculate_cluster_membership(cluster_assignments, rng)
            labels = self._ensure_all_clusters_in_labels(labels, rng)

            if (labels == self.labels_).all():
                break

            self.labels_ = labels
            iterations += 1

    def _initial_cluster_centers_plus(self, cluster_assignments, rng):
        cluster_assignments = cluster_assignments.T

        self._cluster_centers = np.zeros(
            (self.n_clusters, cluster_assignments.shape[1]), dtype=np.int32
        )
        self._cluster_centers[0] = cluster_assignments[
            rng.randint(cluster_assignments.shape[0])
        ]

        for i in range(1, self.n_clusters):
            dists = np.sum(cluster_assignments != self._cluster_centers[i - 1], axis=1)

            probability = dists / dists.sum()
            next_center_index = rng.choice(cluster_assignments.shape[0], p=probability)

            self._cluster_centers[i] = cluster_assignments[next_center_index]

    def _initial_cluster_centers_random(self, cluster_assignments, rng):
        self.labels_ = rng.randint(self.n_clusters, size=cluster_assignments.shape[1])
        self.labels_ = self._ensure_all_clusters_in_labels(self.labels_, rng)

        self._select_cluster_centers(cluster_assignments, rng)

    def _initial_cluster_centers_aligned(self, cluster_assignments, rng):
        new_assignments = np.zeros(
            (cluster_assignments.shape[0], cluster_assignments.shape[1]), dtype=np.int32
        )

        new_assignments[0] = cluster_assignments[0]
        for i in range(1, len(cluster_assignments)):
            cost_matrix = -np.histogram2d(
                cluster_assignments[i], cluster_assignments[0], bins=self.n_clusters
            )[0]
            _, col_indices = linear_sum_assignment(cost_matrix)
            new_assignments[i] = col_indices[cluster_assignments[i]]

        votes = np.apply_along_axis(
            lambda x: np.bincount(x, minlength=self.n_clusters),
            axis=0,
            arr=new_assignments,
        ).T

        self.labels_ = np.array(
            [rng.choice(np.flatnonzero(v == v.max())) for v in votes]
        )
        self.labels_ = self._ensure_all_clusters_in_labels(self.labels_, rng)

        self._select_cluster_centers(cluster_assignments, rng)

    def _select_cluster_centers(self, cluster_assignments, rng):
        self._cluster_centers = np.zeros(
            (self.n_clusters, cluster_assignments.shape[0]), dtype=np.int32
        )

        for i in range(self.n_clusters):
            cluster_indices = np.where(self.labels_ == i)[0]
            p = cluster_assignments[:, cluster_indices]

            for n, att in enumerate(p):
                unique, counts = np.unique(att, return_counts=True)
                self._cluster_centers[i][n] = unique[
                    rng.choice(np.flatnonzero(counts == counts.max()))
                ]

        for i in range(1, self.n_clusters):
            for j in range(i):
                if (self._cluster_centers[i] == self._cluster_centers[j]).all():
                    self._cluster_centers[i] = cluster_assignments[
                        :, rng.randint(cluster_assignments.shape[1])
                    ]

    def _calculate_cluster_membership(self, cluster_assignments, rng):
        labels = np.zeros(cluster_assignments.shape[1], dtype=np.int32)

        for i in range(cluster_assignments.shape[1]):
            dists = np.sum(cluster_assignments[:, i] != self._cluster_centers, axis=1)

            min_indices = np.where(dists == dists.min())[0]

            if len(min_indices) > 1:
                labels[i] = rng.choice(min_indices)
            else:
                labels[i] = min_indices[0]

        return labels

    def _ensure_all_clusters_in_labels(self, labels, rng):
        unique = np.unique(labels)
        if unique.shape[0] != self.n_clusters:
            for i in range(self.n_clusters):
                if i not in unique:
                    x = np.concatenate(
                        [
                            np.where(labels == unique[i])[0]
                            for i in range(unique.shape[0])
                        ]
                    )
                    labels[rng.choice(x)] = i

        return labels



"""Consensus clustering algorithms."""

__all__ = [
    "IterativeVotingClustering",
    "FromFileIterativeVotingClustering",
    "SimpleVote",
    "FromFileSimpleVote",
]

from tsml_eval.estimators.clustering.consensus.ivc import IterativeVotingClustering
from tsml_eval.estimators.clustering.consensus.ivc_from_file import (
    FromFileIterativeVotingClustering,
)
from tsml_eval.estimators.clustering.consensus.simple_vote import SimpleVote
from tsml_eval.estimators.clustering.consensus.simple_vote_from_file import (
    FromFileSimpleVote,
)



import numpy as np
from sklearn.utils.validation import check_random_state

from tsml_eval._wip.tschief._splitters import (
    DictionarySplitter,
    DistanceSplitter,
    IntervalSplitter,
)


class TsChiefNode:
    """Tree node for the TS-CHIEF model."""

    def __init__(self, n_dictionary, n_distance, n_interval, random_state):
        self.rng = check_random_state(random_state)
        self.n_dictionary = n_dictionary
        self.n_distance = n_distance
        self.n_interval = n_interval

        self.is_leaf = False

    def fit(self, X, y, X_boss, sfas):
        """Fit the tree node with respective subset of data."""
        self._class_dtype = y.dtype

        if len(np.unique(y)) == 1:
            self.is_leaf = True
            self.label = y[0]
            return self

        best_splitter = None
        best_split = None
        best_gini = np.inf

        for i in range(self.n_dictionary):
            splitter = DictionarySplitter.generate(X_boss, y, sfas, self.rng)
            split = splitter.split_train(X_boss)
            gini = _gini(y, split)

            if gini < best_gini:
                best_splitter = splitter
                best_split = split
                best_gini = gini

        for i in range(self.n_distance):
            splitter = DistanceSplitter.generate(X, y, self.rng)
            split = splitter.split(X)
            gini = _gini(y, split)

            if gini < best_gini:
                best_splitter = splitter
                best_split = split
                best_gini = gini

        for i in range(self.n_interval):
            splitter = IntervalSplitter.generate(X, y, self.rng)
            split = splitter.split(X)
            gini = _gini(y, split)

            if gini < best_gini:
                best_splitter = splitter
                best_split = split
                best_gini = gini

        self.splitter_ = best_splitter
        self.children_ = []
        for split in np.unique(best_split):
            down_idx = np.argwhere(best_split == split).ravel()
            down_X = X[down_idx]
            down_y = y[down_idx]
            down_X_boss = np.vectorize(lambda bag, down_idx=down_idx: bag[down_idx, :])(
                X_boss
            )

            self.children_.append(
                TsChiefNode(
                    n_dictionary=self.n_dictionary,
                    n_distance=self.n_distance,
                    n_interval=self.n_interval,
                    random_state=self.rng,
                ).fit(down_X, down_y, down_X_boss, sfas)
            )

        return self

    def predict(self, X):
        """Predict class labels for subset of data."""
        samples = X.shape[0]

        if self.is_leaf:
            return np.repeat(self.label, samples)

        preds = np.empty(samples, dtype=self._class_dtype)
        split_idx = self.splitter_.split(X)
        for split in np.unique(split_idx):
            down_idx = np.argwhere(split_idx == split).ravel()
            preds[down_idx] = self.children_[split].predict(X[down_idx])

        return preds


def _gini(y, split_idx):
    ginis = []
    splits, split_sizes = np.unique(split_idx, return_counts=True)
    for split in splits:
        class_distribution = y[np.argwhere(split_idx == split).ravel()]

        _, class_counts = np.unique(class_distribution, return_counts=True)
        class_counts = class_counts / len(class_distribution)
        ginis.append(1 - np.sum(class_counts**2))

    return np.average(ginis, weights=split_sizes)



"""Tests for clustering experiments."""

__author__ = ["MatthewMiddlehurst"]

import os
import runpy

import pytest
from aeon.registry import all_estimators
from tsml.dummy import DummyClassifier, DummyClusterer

from tsml_eval.datasets._test_data._data_sizes import DATA_TEST_SIZES, DATA_TRAIN_SIZES
from tsml_eval.experiments import (
    clustering_experiments,
    get_clusterer_by_name,
    run_clustering_experiment,
    set_clusterer,
    threaded_clustering_experiments,
)
from tsml_eval.experiments.tests import _CLUSTERER_RESULTS_PATH
from tsml_eval.testing.testing_utils import (
    _TEST_DATA_PATH,
    _check_set_method,
    _check_set_method_results,
)
from tsml_eval.utils.tests.test_results_writing import _check_clustering_file_format


@pytest.mark.parametrize(
    "clusterer",
    ["DummyClusterer-tsml", "DummyClusterer-aeon", "DummyClusterer-sklearn"],
)
@pytest.mark.parametrize(
    "dataset",
    ["MinimalChinatown", "UnequalMinimalChinatown", "EqualMinimalJapaneseVowels"],
)
def test_run_clustering_experiment(clusterer, dataset):
    """Test clustering experiments with test data and clusterer."""
    if clusterer == "DummyClusterer-aeon" and dataset == "UnequalMinimalChinatown":
        return  # todo remove when aeon kmeans supports unequal

    args = [
        _TEST_DATA_PATH,
        _CLUSTERER_RESULTS_PATH,
        clusterer,
        dataset,
        "0",
        "-te",
    ]

    clustering_experiments.run_experiment(args)

    test_file = (
        f"{_CLUSTERER_RESULTS_PATH}{clusterer}/Predictions/{dataset}/testResample0.csv"
    )
    train_file = (
        f"{_CLUSTERER_RESULTS_PATH}{clusterer}/Predictions/{dataset}/trainResample0.csv"
    )

    assert os.path.exists(test_file) and os.path.exists(train_file)
    _check_clustering_file_format(test_file, num_results_lines=DATA_TEST_SIZES[dataset])
    _check_clustering_file_format(
        train_file, num_results_lines=DATA_TEST_SIZES[dataset]
    )

    # test present results checking
    clustering_experiments.run_experiment(args)

    os.remove(test_file)
    os.remove(train_file)


def test_run_clustering_experiment_main():
    """Test clustering experiments main with test data and clusterer."""
    clusterer = "KMeans"
    dataset = "MinimalChinatown"

    # run twice to test results present check
    for _ in range(2):
        runpy.run_path(
            (
                "./tsml_eval/experiments/clustering_experiments.py"
                if os.getcwd().split("\\")[-1] != "tests"
                else "../clustering_experiments.py"
            ),
            run_name="__main__",
        )

    train_file = (
        f"{_CLUSTERER_RESULTS_PATH}{clusterer}/Predictions/{dataset}/trainResample0.csv"
    )
    assert os.path.exists(train_file)
    _check_clustering_file_format(train_file)

    os.remove(train_file)


def test_run_threaded_clustering_experiment():
    """Test threaded clustering experiments with test data and clusterer."""
    clusterer = "KMeans"
    dataset = "MinimalChinatown"

    args = [
        _TEST_DATA_PATH,
        _CLUSTERER_RESULTS_PATH,
        clusterer,
        dataset,
        "1",
        "-nj",
        "2",
        # also test normalisation and benchmark time here
        "--row_normalise",
        "--benchmark_time",
        "-te",
    ]

    threaded_clustering_experiments.run_experiment(args)

    train_file = (
        f"{_CLUSTERER_RESULTS_PATH}{clusterer}/Predictions/{dataset}/trainResample1.csv"
    )
    test_file = (
        f"{_CLUSTERER_RESULTS_PATH}{clusterer}/Predictions/{dataset}/testResample1.csv"
    )

    assert os.path.exists(train_file) and os.path.exists(test_file)

    _check_clustering_file_format(train_file)
    _check_clustering_file_format(test_file)

    # test present results checking
    threaded_clustering_experiments.run_experiment(args)

    # this covers the main method and experiment function result file checking
    runpy.run_path(
        (
            "./tsml_eval/experiments/threaded_clustering_experiments.py"
            if os.getcwd().split("\\")[-1] != "tests"
            else "../threaded_clustering_experiments.py"
        ),
        run_name="__main__",
    )

    os.remove(train_file)
    os.remove(test_file)


def test_run_clustering_experiment_invalid_build_settings():
    """Test run_clustering_experiment method with invalid build settings."""
    with pytest.raises(ValueError, match="Both test_file and train_file"):
        run_clustering_experiment(
            [],
            [],
            None,
            "",
            build_train_file=False,
            build_test_file=False,
        )


def test_run_clustering_experiment_invalid_estimator():
    """Test run_clustering_experiment method with invalid estimator."""
    with pytest.raises(TypeError, match="clusterer must be a"):
        run_clustering_experiment(
            [],
            [],
            DummyClassifier(),
            "",
        )


def test_get_clusterer_by_name():
    """Test get_clusterer_by_name method."""
    clusterer_lists = [
        set_clusterer.distance_based_clusterers,
        set_clusterer.other_clusterers,
        set_clusterer.vector_clusterers,
    ]

    clusterer_dict = {}
    all_clusterer_names = []

    for clusterer_list in clusterer_lists:
        _check_set_method(
            get_clusterer_by_name,
            clusterer_list,
            clusterer_dict,
            all_clusterer_names,
        )

    _check_set_method_results(
        clusterer_dict, estimator_name="Clusterers", method_name="set_clusterer"
    )


def test_get_clusterer_by_name_invalid():
    """Test get_clusterer_by_name method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN CLUSTERER"):
        get_clusterer_by_name("invalid")


def test_aeon_clusterers_available():
    """Test all aeon clusterers are available."""
    excluded = [
        # composable
        "ClustererPipeline",
        # just missing
        "AEFCNClusterer",
        "AEResNetClusterer",
        "TimeSeriesKShapes",
        "TimeSeriesKernelKMeans",
    ]

    est = [e for e, _ in all_estimators(estimator_types="clusterer")]
    for e in est:
        if e in excluded:
            continue

        try:
            assert get_clusterer_by_name(e) is not None
        except ModuleNotFoundError:
            continue


@pytest.mark.parametrize("n_clusters", ["4", "-1"])
@pytest.mark.parametrize(
    "clusterer",
    ["DBSCAN", "DummyClusterer-aeon", "DummyClusterer-sklearn"],
)
def test_n_clusters(n_clusters, clusterer):
    """Test n_clusters parameter."""
    dataset = "MinimalChinatown"

    args = [
        _TEST_DATA_PATH,
        _CLUSTERER_RESULTS_PATH + f"{n_clusters}/",
        clusterer,
        dataset,
        "1",
        "--n_clusters",
        n_clusters,
        "-ow",
    ]

    clustering_experiments.run_experiment(args)

    train_file = (
        f"{_CLUSTERER_RESULTS_PATH}{n_clusters}/{clusterer}/Predictions/{dataset}/"
        "trainResample1.csv"
    )

    assert os.path.exists(train_file)

    if clusterer != "DBSCAN":
        _check_clustering_file_n_clusters(
            train_file, "2" if n_clusters == "-1" else n_clusters
        )

    os.remove(train_file)


@pytest.mark.parametrize(
    "dataset",
    ["MinimalChinatown", "UnequalMinimalChinatown", "EqualMinimalJapaneseVowels"],
)
def test_combined_train_test(dataset):
    """Test n_clusters parameter."""
    clusterer = "DummyClusterer-tsml"

    args = [
        _TEST_DATA_PATH,
        _CLUSTERER_RESULTS_PATH + "Combined/",
        clusterer,
        dataset,
        "1",
        "-ow",
        "-ctts",
    ]

    clustering_experiments.run_experiment(args)

    train_file = (
        f"{_CLUSTERER_RESULTS_PATH}Combined/{clusterer}/Predictions/{dataset}/"
        "trainResample1.csv"
    )

    assert os.path.exists(train_file)
    _check_clustering_file_format(
        train_file,
        num_results_lines=DATA_TRAIN_SIZES[dataset] + DATA_TEST_SIZES[dataset],
    )

    os.remove(train_file)


def _check_clustering_file_n_clusters(file_path, expected):
    with open(file_path) as f:
        lines = f.readlines()

    line = lines[2].split(",")
    assert line[6].strip() == expected


def test_invalid_n_clusters():
    """Test invalid n_clusters parameter."""
    with pytest.raises(ValueError, match="n_clusters must be a"):
        run_clustering_experiment(
            [],
            [],
            DummyClusterer(),
            "",
            n_clusters="invalid",
        )


def test_invalid_test_settings():
    """Test build_test_file without test data."""
    with pytest.raises(ValueError, match="Test data and labels not provided"):
        run_clustering_experiment(
            [],
            [],
            DummyClusterer(),
            "",
            build_test_file=True,
        )



"""Tests for cross-validation functions."""

import os

import numpy as np
from aeon.classification import DummyClassifier
from aeon.datasets import load_covid_3month, load_unit_test
from aeon.regression import DummyRegressor

from tsml_eval.experiments.cross_validation import (
    classification_cross_validation,
    classification_cross_validation_folds,
    regression_cross_validation,
    regression_cross_validation_folds,
)
from tsml_eval.experiments.tests import (
    _CLASSIFIER_RESULTS_PATH,
    _REGRESSOR_RESULTS_PATH,
)
from tsml_eval.utils.tests.test_results_writing import (
    _check_classification_file_format,
    _check_regression_file_format,
)


def test_classification_cross_validation():
    """Test the classification cross-validation function."""
    X, y = load_unit_test()
    classification_cross_validation(
        X,
        y,
        DummyClassifier(),
        _CLASSIFIER_RESULTS_PATH,
        classifier_name="DummyClassifierCV",
    )

    for i in range(10):
        test_file = (
            f"{_CLASSIFIER_RESULTS_PATH}/DummyClassifierCV/Predictions/"
            f"testResample{i}.csv"
        )

        assert os.path.exists(test_file)
        _check_classification_file_format(test_file)

    folds = classification_cross_validation_folds(X, y)
    assert len(folds) == 10
    assert len(folds[0]) == 2
    assert isinstance(folds[0][0], np.ndarray)
    assert isinstance(folds[0][1], np.ndarray)


def test_regression_cross_validation():
    """Test the classification cross-validation function."""
    X, y = load_covid_3month()
    regression_cross_validation(
        X,
        y,
        DummyRegressor(),
        _REGRESSOR_RESULTS_PATH,
        regressor_name="DummyRegressorCV",
    )

    for i in range(10):
        test_file = (
            f"{_REGRESSOR_RESULTS_PATH}/DummyRegressorCV/Predictions/"
            f"testResample{i}.csv"
        )

        assert os.path.exists(test_file)
        _check_regression_file_format(test_file)

    folds = regression_cross_validation_folds(X, y)
    assert len(folds) == 10
    assert len(folds[0]) == 2
    assert isinstance(folds[0][0], np.ndarray)
    assert isinstance(folds[0][1], np.ndarray)



"""Clusterers used in the publication."""

__author__ = ["MatthewMiddlehurst"]

from tsml_eval.utils.functions import str_in_nested_list

distance_based_clusterers = [
    ["TimeSeriesKMeans", "kmeans-dtw", "k-means-dtw"],
    ["kmeans-ddtw", "k-means-ddtw"],
    ["kmeans-ed", "k-means-ed"],
    ["kmeans-edr", "k-means-edr"],
    ["kmeans-erp", "k-means-erp"],
    ["kmeans-lcss", "k-means-lcss"],
    ["kmeans-msm", "k-means-msm"],
    ["kmeans-twe", "k-means-twe"],
    ["kmeans-wdtw", "k-means-wdtw"],
    ["kmeans-wddtw", "k-means-wddtw"],
    ["TimeSeriesKMedoids", "kmedoids-dtw", "k-medoids-dtw"],
    ["kmedoids-ddtw", "k-medoids-ddtw"],
    ["kmedoids-ed", "k-medoids-ed"],
    ["kmedoids-edr", "k-medoids-edr"],
    ["kmedoids-erp", "k-medoids-erp"],
    ["kmedoids-lcss", "k-medoids-lcss"],
    ["kmedoids-msm", "k-medoids-msm"],
    ["kmedoids-twe", "k-medoids-twe"],
    ["kmedoids-wdtw", "k-medoids-wdtw"],
    ["kmedoids-wddtw", "k-medoids-wddtw"],
]


def _set_distance_clusterer(
    clusterer_name,
    random_state=None,
    **kwargs,
):
    c = clusterer_name.lower()

    if not str_in_nested_list(distance_based_clusterers, c):
        raise ValueError(f"UNKNOWN CLUSTERER: {c} in set_distance_clusterer")

    if c == "timeserieskmeans" or c == "kmeans-dtw" or c == "k-means-dtw":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="dtw", random_state=random_state, **kwargs)
    elif c == "kmeans-ddtw" or c == "k-means-ddtw":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="ddtw", random_state=random_state, **kwargs)
    elif c == "kmeans-ed" or c == "k-means-ed":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(
            distance="euclidean", random_state=random_state, **kwargs
        )
    elif c == "kmeans-edr" or c == "k-means-edr":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="edr", random_state=random_state, **kwargs)
    elif c == "kmeans-erp" or c == "k-means-erp":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="erp", random_state=random_state, **kwargs)
    elif c == "kmeans-lcss" or c == "k-means-lcss":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="lcss", random_state=random_state, **kwargs)
    elif c == "kmeans-msm" or c == "k-means-msm":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="msm", random_state=random_state, **kwargs)
    elif c == "kmeans-twe" or c == "k-means-twe":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="twe", random_state=random_state, **kwargs)
    elif c == "kmeans-wdtw" or c == "k-means-wdtw":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="wdtw", random_state=random_state, **kwargs)
    elif c == "kmeans-wddtw" or c == "k-means-wddtw":
        from aeon.clustering import TimeSeriesKMeans

        return TimeSeriesKMeans(distance="wddtw", random_state=random_state, **kwargs)
    elif c == "timeserieskmedoids" or c == "kmedoids-dtw" or c == "k-medoids-dtw":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="dtw", random_state=random_state, **kwargs)
    elif c == "kmedoids-ddtw" or c == "k-medoids-ddtw":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="ddtw", random_state=random_state, **kwargs)
    elif c == "kmedoids-ed" or c == "k-medoids-ed":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(
            distance="euclidean", random_state=random_state, **kwargs
        )
    elif c == "kmedoids-edr" or c == "k-medoids-edr":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="edr", random_state=random_state, **kwargs)
    elif c == "kmedoids-erp" or c == "k-medoids-erp":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="erp", random_state=random_state, **kwargs)
    elif c == "kmedoids-lcss" or c == "k-medoids-lcss":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="lcss", random_state=random_state, **kwargs)
    elif c == "kmedoids-msm" or c == "k-medoids-msm":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="msm", random_state=random_state, **kwargs)
    elif c == "kmedoids-twe" or c == "k-medoids-twe":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="twe", random_state=random_state, **kwargs)
    elif c == "kmedoids-wdtw" or c == "k-medoids-wdtw":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="wdtw", random_state=random_state, **kwargs)
    elif c == "kmedoids-wddtw" or c == "k-medoids-wddtw":
        from aeon.clustering import TimeSeriesKMedoids

        return TimeSeriesKMedoids(distance="wddtw", random_state=random_state, **kwargs)



"""Code for tuning clusterer parameters used in the publication."""

import sys

import numpy as np
from aeon.clustering import TimeSeriesKMeans
from sklearn.metrics import davies_bouldin_score


# used for dtw and wdtw primarily
def _tune_window(metric, train_X, n_clusters):  # pragma: no cover
    best_w = 0
    best_score = sys.float_info.max
    for w in np.arange(0.0, 0.2, 0.01):
        cls = TimeSeriesKMeans(
            metric=metric, distance_params={"window": w}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(f" Number of clusters = {clusters} window = {w} score = {score}")
        if score < best_score:
            best_score = score
            best_w = w
    print("best window =", best_w, " with score ", best_score)
    return best_w


def _tune_msm(train_X, n_clusters):  # pragma: no cover
    best_c = 0
    best_score = sys.float_info.max
    for c in np.arange(0.0, 5.0, 0.25):
        cls = TimeSeriesKMeans(
            metric="msm", distance_params={"c": c}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(f" Number of clusters = {clusters} c parameter = {c} score = {score}")
        if score < best_score:
            best_score = score
            best_c = c
    print("best c =", best_c, " with score ", best_score)
    return best_c


def _tune_wdtw(train_X, n_clusters):  # pragma: no cover
    best_g = 0
    best_score = sys.float_info.max
    for g in np.arange(0.0, 1.0, 0.05):
        cls = TimeSeriesKMeans(
            metric="wdtw", distance_params={"g": g}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(f" Number of clusters = {clusters} g parameter = {g} score = {score}")
        if score < best_score:
            best_score = score
            best_g = g
    print("best g =", best_g, " with score ", best_score)
    return best_g


def _tune_twe(train_X, n_clusters):  # pragma: no cover
    best_nu = 0
    best_lambda = 0
    best_score = sys.float_info.max
    for nu in np.arange(0.0, 1.0, 0.25):
        for lam in np.arange(0.0, 1.0, 0.2):
            cls = TimeSeriesKMeans(
                metric="twe",
                distance_params={"nu": nu, "lmbda": lam},
                n_clusters=n_clusters,
            )
            cls.fit(train_X)
            preds = cls.predict(train_X)
            clusters = len(np.unique(preds))
            if clusters <= 1:
                score = sys.float_info.max
            else:
                score = davies_bouldin_score(train_X, preds)
            print(
                f" Number of clusters = {clusters} nu param = {nu} lambda para "
                f"= {lam} score = {score}"
            )  #
            # noqa
            if score < best_score:
                best_score = score
                best_nu = nu
                best_lambda = lam
    print("best nu =", best_nu, f" lambda = {best_lambda} score ", best_score)  # noqa
    return best_nu, best_lambda


def _tune_erp(train_X, n_clusters):  # pragma: no cover
    best_g = 0
    best_score = sys.float_info.max
    for g in np.arange(0.0, 2.0, 0.2):
        cls = TimeSeriesKMeans(
            metric="erp", distance_params={"g": g}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(f" Number of clusters ={clusters} g parameter = {g} score  = {score}")
        if score < best_score:
            best_score = score
            best_g = g
    print("best g =", best_g, " with score ", best_score)
    return best_g


def _tune_edr(train_X, n_clusters):  # pragma: no cover
    best_e = 0
    best_score = sys.float_info.max
    for e in np.arange(0.0, 0.2, 0.01):
        cls = TimeSeriesKMeans(
            metric="edr", distance_params={"epsilon": e}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(
            f" Number of clusters = {clusters} epsilon parameter = {e} score = {score}"
        )
        if score < best_score:
            best_score = score
            best_e = e
    print("best e =", best_e, " with score ", best_score)  # noqa
    return best_e


def _tune_lcss(train_X, n_clusters):  # pragma: no cover
    best_e = 0
    best_score = sys.float_info.max
    for e in np.arange(0.0, 0.2, 0.01):
        cls = TimeSeriesKMeans(
            metric="lcss", distance_params={"epsilon": e}, n_clusters=n_clusters
        )
        cls.fit(train_X)
        preds = cls.predict(train_X)
        clusters = len(np.unique(preds))
        if clusters <= 1:
            score = sys.float_info.max
        else:
            score = davies_bouldin_score(train_X, preds)
        print(
            f" Number of clusters ={clusters} epsilon parameter = {e} score = {score}"
        )
        if score < best_score:
            best_score = score
            best_e = e
    print("best e =", best_e, " with score ", best_score)
    return best_e



"""Tests for publication experiments estimator selection."""

import pytest

from tsml_eval.publications.y2023.tsc_bakeoff import (
    _set_bakeoff_classifier,
    bakeoff_classifiers,
)
from tsml_eval.testing.testing_utils import _check_set_method, _check_set_method_results


def test_set_bakeoff_classifier():
    """Test set_bakeoff_classifier method."""
    classifier_dict = {}
    all_classifier_names = []

    _check_set_method(
        _set_bakeoff_classifier,
        bakeoff_classifiers,
        classifier_dict,
        all_classifier_names,
    )

    _check_set_method_results(
        classifier_dict,
        estimator_name="Classifiers",
        method_name="_set_bakeoff_classifier",
    )


def test_set_bakeoff_classifier_invalid():
    """Test set_bakeoff_classifier method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN CLASSIFIER"):
        _set_bakeoff_classifier("invalid")



"""RIST pipeline publication tests."""

__all__ = ["_RIST_TEST_RESULTS_PATH"]

import os
from pathlib import Path

_RIST_TEST_RESULTS_PATH = (
    os.path.dirname(Path(__file__).parent.parent.parent.parent.parent)
    + "/test_output/rist_pipeline/"
)
